from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..exceptions import PlotnineWarning
from ..utils import remove_missing
from .position import position

if typing.TYPE_CHECKING:
    from plotnine.typing import Scale, Trans


class position_stack(position):
    """
    Stack plotted objects on top of each other

    The objects to stack are those that have
    an overlapping x range.
    """

    fill = False

    def __init__(self, vjust=1, reverse=False):
        self.params = {"vjust": vjust, "reverse": reverse}

    def setup_params(self, data):
        """
        Verify, modify & return a copy of the params.
        """
        # Variable for which to do the stacking
        if "ymax" in data:
            if any((data["ymin"] != 0) & (data["ymax"] != 0)):
                warn(
                    "Stacking not well defined when not "
                    "anchored on the axis.",
                    PlotnineWarning,
                )
            var = "ymax"
        elif "y" in data:
            var = "y"
        else:
            warn(
                "Stacking requires either ymin & ymax or y "
                "aesthetics. Maybe you want position = 'identity'?",
                PlotnineWarning,
            )
            var = None

        params = self.params.copy()
        params["var"] = var
        params["fill"] = self.fill
        return params

    def setup_data(self, data, params):
        if not params["var"]:
            return data

        if params["var"] == "y":
            data["ymax"] = data["y"]
        elif params["var"] == "ymax":
            bool_idx = data["ymax"] == 0
            data.loc[bool_idx, "ymax"] = data.loc[bool_idx, "ymin"]

        data = remove_missing(
            data, vars=("x", "xmin", "xmax", "y"), name="position_stack"
        )

        return data

    @classmethod
    def compute_panel(cls, data, scales, params):
        if not params["var"]:
            return data

        # TODO: Make transforms in mizani aware of their
        # linear status
        def _is_non_linear_trans(trans: Trans) -> bool:
            tname = trans.__class__.__name__
            linear_transforms = ("identity", "reverse")
            if tname.endswith("_trans"):
                tname = tname[:-6]
            return (
                trans.dataspace_is_numerical and tname not in linear_transforms
            )

        def get_non_linear_trans(sc: Scale) -> Trans | None:
            """
            Return trans if the scale in non-linear
            """
            from ..scales.scale import scale_continuous

            if isinstance(sc, scale_continuous):
                if _is_non_linear_trans(sc.trans):
                    return sc.trans

        # Positioning happens after scale transformation and stacking
        # only works well for linear data. If the scale is non-linear,
        # we undo the transformation, stack and redo the transformation
        nl_trans = get_non_linear_trans(scales.y)
        if nl_trans:
            data = cls.transform_position(data, trans_y=nl_trans.inverse)

        negative = data["ymax"] < 0
        neg = data.loc[negative]
        pos = data.loc[~negative]

        if len(neg):
            neg = cls.collide(neg, params=params)

        if len(pos):
            pos = cls.collide(pos, params=params)

        data = pd.concat([neg, pos], axis=0, ignore_index=True, sort=True)

        if nl_trans:
            data = cls.transform_position(data, trans_y=nl_trans.transform)

        return data

    @staticmethod
    def strategy(data, params):
        """
        Stack overlapping intervals.

        Assumes that each set has the same horizontal position
        """
        vjust = params["vjust"]

        y = data["y"].copy()
        y[np.isnan(y)] = 0
        heights = np.append(0, y.cumsum())

        if params["fill"]:
            heights = heights / np.abs(heights[-1])

        data["ymin"] = np.min([heights[:-1], heights[1:]], axis=0)
        data["ymax"] = np.max([heights[:-1], heights[1:]], axis=0)
        # less intuitive than (ymin + vjust(ymax-ymin)), but
        # this way avoids subtracting numbers of potentially
        # similar precision
        data["y"] = (1 - vjust) * data["ymin"] + vjust * data["ymax"]
        return data



import numpy as np
import pandas as pd

from plotnine import aes, geom_bin_2d, ggplot, scale_x_log10

from .conftest import layer_data

n = 20  # Make even for best results
reps = np.hstack(
    [np.arange(int(np.ceil(n / 2))), np.arange(int(np.ceil(n // 2)))[::-1]]
)
diagonal = np.repeat(np.arange(n), reps)

df = pd.DataFrame(
    {
        "x": np.hstack([diagonal, diagonal]),
        "y": np.hstack([diagonal, diagonal[::-1]]),
    }
)


def test_drop_true():
    p = ggplot(df, aes("x", "y")) + geom_bin_2d(binwidth=2, drop=True)
    assert p == "drop_true"


def test_drop_false():
    p = ggplot(df, aes("x", "y")) + geom_bin_2d(binwidth=2, drop=False)
    assert p == "drop_false"


def test_scale_transformed_breaks():
    df = pd.DataFrame({"x": [1, 10, 100, 1000], "y": range(4)})
    p = ggplot(df, aes("x", "y")) + geom_bin_2d(
        breaks=([5, 50, 500], [0.5, 1.5, 2.5])
    )
    out1 = layer_data(p)
    out2 = layer_data(p + scale_x_log10())
    np.testing.assert_allclose(out1.xmax, [50, 500])
    np.testing.assert_allclose(out2.xmax, np.log10([50, 500]))



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, geom_point, ggplot, stat_summary

random_state = np.random.RandomState(1234567890)

df = pd.DataFrame(
    {
        "x": list("aaaaabbbbcccccc"),
        "y": [1, 2, 3, 4, 5, 1.5, 1.5, 6, 6, 5, 5, 5, 5, 5, 5],
    }
)


def test_mean_cl_boot():
    p = ggplot(df, aes("x", "y")) + stat_summary(
        fun_data="mean_cl_boot", random_state=random_state, size=2
    )

    assert p == "mean_cl_boot"


def test_mean_cl_normal():
    p = ggplot(df, aes("x", "y")) + stat_summary(
        fun_data="mean_cl_normal", size=2
    )

    assert p == "mean_cl_normal"


def test_mean_sdl():
    p = ggplot(df, aes("x", "y")) + stat_summary(fun_data="mean_sdl", size=2)

    assert p == "mean_sdl"


def test_median_hilow():
    p = ggplot(df, aes("x", "y")) + stat_summary(
        fun_data="median_hilow", size=2
    )

    assert p == "median_hilow"


def test_mean_se():
    p = ggplot(df, aes("x", "y")) + stat_summary(fun_data="mean_se", size=2)

    assert p == "mean_se"


def test_funargs():
    p = (
        ggplot(df, aes("x", "y"))
        + stat_summary(fun_data="mean_cl_normal", size=2, color="blue")
        + stat_summary(
            fun_data="mean_cl_normal",
            fun_args={"confidence_interval": 0.5},
            size=2,
            color="green",
        )
    )

    assert p == "fun_args"


def test_summary_functions():
    p = ggplot(df, aes("x", "y")) + stat_summary(
        fun_y=np.mean, fun_ymin=np.min, fun_ymax=np.max, size=2
    )

    assert p == "summary_functions"


def test_stat_summary_raises_on_invalid_paremeters():
    with pytest.raises(TypeError):
        geom_point(stat_summary(funy=np.mean))
    with pytest.raises(TypeError):
        geom_point(stat_summary(does_not_exist=1))



from copy import deepcopy

import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    after_scale,
    after_stat,
    annotate,
    coord_trans,
    facet_null,
    geom_bar,
    geom_histogram,
    geom_line,
    geom_point,
    ggplot,
    ggtitle,
    guides,
    labs,
    lims,
    scale_x_continuous,
    stage,
    stat_identity,
    theme,
    theme_gray,
    xlab,
    xlim,
    ylab,
)
from plotnine.exceptions import PlotnineError, PlotnineWarning
from plotnine.mapping.aes import is_valid_aesthetic

df = pd.DataFrame({"x": np.arange(10), "y": np.arange(10)})


def test_labels():
    """
    Test invalid arguments to chart components
    """
    gg = ggplot(df, aes(x="x", y="y"))
    gg = gg + geom_point()
    gg = gg + xlab("xlab")
    gg = gg + ylab("ylab")
    gg = gg + ggtitle("title")

    assert gg.labels.x == "xlab"
    assert gg.labels.y == "ylab"
    assert gg.labels.title == "title"

    gg = gg + labs(x="xlab2", y="ylab2", title="title2", caption="caption2")
    assert gg.labels.x == "xlab2"
    assert gg.labels.y == "ylab2"
    assert gg.labels.title == "title2"
    assert gg.labels.caption == "caption2"

    with pytest.raises(PlotnineError):
        gg = gg + labs(z="z-axis")

    with pytest.raises(PlotnineError):
        gg = gg + xlab(None)

    with pytest.raises(PlotnineError):
        gg = gg + ylab(None)

    with pytest.raises(PlotnineError):
        gg = gg + ggtitle(None)


def test_ggplot_parameters():
    p = ggplot(df, aes("x"))

    assert p.data is df
    assert p.mapping == aes("x")
    assert p.environment.namespace["np"] is np
    assert p.environment.namespace["pd"] is pd

    p = ggplot(data=df, mapping=aes("x"))
    assert p.data is df
    assert p.mapping == aes("x")

    p = ggplot(data=df)
    assert p.data is df
    assert p.mapping == aes()

    p = ggplot(mapping=aes("x"))
    assert p.data is None
    assert p.mapping == aes("x")

    p = ggplot()
    assert p.data is None
    assert p.mapping == aes()

    with pytest.raises(TypeError):
        ggplot([1, 2, 3], aes("x"))


def test_ggplot_parameters_grouped():
    p = df.groupby("x") >> ggplot(aes("x"))
    assert isinstance(p.data, pd.DataFrame)


def test_data_transforms():
    p = ggplot(aes(x="x", y="np.log(y+1)"), df)
    p = p + geom_point()
    p.draw_test()

    with pytest.raises(Exception):
        # no numpy available
        p = ggplot(aes(x="depth", y="ap.log(price)"), df)
        p = p + geom_point()
        p.draw_test()


def test_deepcopy():
    p = ggplot(df, aes("x")) + geom_histogram()
    p2 = deepcopy(p)
    assert p is not p2
    # Not sure what we have to do for that...
    assert p.data is p2.data
    assert len(p.layers) == len(p2.layers)
    assert p.layers[0].geom is not p2.layers[0].geom
    assert len(p.mapping) == len(p2.mapping)
    assert p.mapping is not p2.mapping
    assert p.environment is p2.environment


def test_aes():
    result = aes("weight", "hp", color="qsec")
    expected = {"x": "weight", "y": "hp", "color": "qsec"}
    assert result == expected

    mapping = aes("weight", "hp", color=stage("qsec"))
    assert mapping["color"].start == "qsec"
    assert mapping._starting["color"] == "qsec"


def test_valid_aes_linetypes():
    assert is_valid_aesthetic("solid", "linetype")
    assert is_valid_aesthetic("--", "linetype")
    assert not is_valid_aesthetic("tada", "linetype")
    assert is_valid_aesthetic((0, (3, 2)), "linetype")
    assert not is_valid_aesthetic((0, (3, 2.0)), "linetype")
    assert not is_valid_aesthetic((0, (3, 2, 1)), "linetype")


def test_valid_aes_shapes():
    assert is_valid_aesthetic("o", "shape")
    assert is_valid_aesthetic((4, 1, 45), "shape")
    assert not is_valid_aesthetic([4, 1, 45], "shape")


def test_valid_aes_colors():
    assert is_valid_aesthetic("red", "color")
    assert is_valid_aesthetic("#FF0000", "color")
    assert is_valid_aesthetic("#FF000080", "color")
    assert is_valid_aesthetic((1, 0, 0), "color")
    assert is_valid_aesthetic((1, 0, 0), "color")
    assert is_valid_aesthetic((1, 0, 0, 0.5), "color")


def test_calculated_aes():
    # after_stat('ae')
    mapping1 = aes("x", y=after_stat("density"))
    mapping2 = aes("x", y=after_stat("density*2"))
    mapping3 = aes("x", y=after_stat("density + count"))
    mapping4 = aes("x", y=after_stat("func(density)"))

    def _test():
        assert list(mapping1._calculated.keys()) == ["y"]
        assert list(mapping2._calculated.keys()) == ["y"]
        assert list(mapping3._calculated.keys()) == ["y"]
        assert list(mapping4._calculated.keys()) == ["y"]

        assert mapping1["y"].after_stat == "density"
        assert mapping2["y"].after_stat == "density*2"
        assert mapping3["y"].after_stat == "density + count"
        assert mapping4["y"].after_stat == "func(density)"

        assert mapping1._calculated["y"] == "density"
        assert mapping2._calculated["y"] == "density*2"
        assert mapping3._calculated["y"] == "density + count"
        assert mapping4._calculated["y"] == "func(density)"

    _test()

    # 'stat(ae)', DEPRECATED but still works
    mapping1 = aes("x", y="stat(density)")
    mapping2 = aes("x", y="stat(density*2)")
    mapping3 = aes("x", y="stat(density + count)")
    mapping4 = aes("x", y="stat(func(density))")
    _test()

    # '..ae..', DEPRECATED but still works
    mapping1 = aes("x", y="..density..")
    mapping2 = aes("x", y="..density..*2")
    mapping3 = aes("x", y="..density.. + ..count..")
    mapping4 = aes("x", y="func(..density..)")
    _test()

    df = pd.DataFrame({"x": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]})
    p = ggplot(df) + geom_bar(aes(x="x", fill=after_stat("count + 2")))
    p.draw_test()

    p = ggplot(df) + geom_bar(aes(x="x", fill="stat(count + 2)"))
    p.draw_test()

    p = ggplot(df) + geom_bar(aes(x="x", fill="..count.. + 2"))
    p.draw_test()


def test_after_scale_mapping():
    df = pd.DataFrame({"x": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]})
    df2 = pd.DataFrame(
        {
            # Same as above, but add 2 of each unique element
            "x": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]
            + [1, 2, 3, 4] * 2
        }
    )

    p = ggplot(df) + geom_bar(aes(x="x", ymax=after_scale("ymax + 2")))
    p2 = ggplot(df2) + geom_bar(aes(x="x"))

    assert p + lims(y=(0, 7)) == "after_scale_mapping"
    assert p2 + lims(y=(0, 7)) == "after_scale_mapping"


def test_add_aes():
    df = pd.DataFrame({"var1": [1, 2, 3, 4], "var2": 2})
    p = ggplot(df) + geom_point()
    p += aes("var1", "var2")

    assert p.mapping == aes("var1", "var2")
    assert p.labels.x == "var1"
    assert p.labels.y == "var2"


def test_nonzero_indexed_data():
    df = pd.DataFrame(
        {98: {"blip": 0, "blop": 1}, 99: {"blip": 1, "blop": 3}}
    ).T
    p = ggplot(df, aes(x="blip", y="blop")) + geom_line()
    p.draw_test()


def test_inplace_add():
    p = _p = ggplot(df)

    p += aes("x", "y")
    assert p is _p

    p += geom_point()
    assert p is _p

    p += stat_identity()
    assert p is _p

    p += scale_x_continuous()
    assert p is _p

    with pytest.warns(PlotnineWarning):
        # Warning for; replacing existing scale added above
        p += xlim(0, 10)
        assert p is _p

    p += lims(y=(0, 10))
    assert p is _p

    p += labs(x="x")
    assert p is _p

    p += coord_trans()
    assert p is _p

    p += facet_null()
    assert p is _p

    p += annotate("point", 5, 5, color="red", size=5)
    assert p is _p

    p += guides()
    assert p is _p

    p += theme_gray()
    assert p is _p

    th = _th = theme_gray()
    th += theme(aspect_ratio=1)
    assert th is _th


def test_rrshift_piping():
    p = df >> ggplot(aes("x", "y")) + geom_point()
    assert p.data is df

    with pytest.raises(PlotnineError):
        df >> ggplot(df.copy(), aes("x", "y")) + geom_point()

    with pytest.raises(TypeError):
        "not a dataframe" >> ggplot(aes("x", "y")) + geom_point()


def test_rrshift_piping_grouped():
    p = df.groupby("x") >> ggplot(aes("x", "y")) + geom_point()
    assert p.data is df


def test_adding_list_ggplot():
    lst = [
        geom_point(),
        geom_point(aes("x+1", "y+1")),
        xlab("x-label"),
        coord_trans(),
    ]
    g = ggplot() + lst
    assert len(g.layers) == 2
    assert g.labels.x == "x-label"
    assert isinstance(g.coordinates, coord_trans)


def test_iadding_list_ggplot():
    lst = [
        geom_point(),
        geom_point(aes("x+1", "y+1")),
        xlab("x-label"),
        coord_trans(),
    ]
    g = ggplot()
    id_before = id(g)
    g += lst
    id_after = id(g)
    assert id_before == id_after
    assert len(g.layers) == 2
    assert g.labels.x == "x-label"
    assert isinstance(g.coordinates, coord_trans)


def test_adding_None():
    p = ggplot(df, aes("x", "y")) + geom_point()
    p2 = p + None
    assert p2 is not p
    assert isinstance(p2, ggplot)

    # Inplace addition
    p += None
    assert isinstance(p, ggplot)


def test_string_group():
    p = ggplot(df, aes("x", "y")) + geom_point(group="pi")
    p.draw_test()


def test_to_pandas():
    class SomeDataType:
        def to_pandas(self):
            return pd.DataFrame({"x": [1, 2, 3], "y": [1, 2, 3]})

    df = SomeDataType()
    p1 = ggplot(df, aes("x", "y")) + geom_point()
    p2 = df >> ggplot(aes("x", "y")) + geom_point()
    assert p1 == "to_pandas"
    assert p2 == "to_pandas"


def test_callable_as_data():
    def _fn(df):
        return df.rename(columns={"xx": "x", "yy": "y"})

    df = pd.DataFrame({"xx": [1, 2, 3], "yy": [1, 2, 3]})
    p = ggplot(df, aes("x", "y")) + geom_point(_fn)
    p.draw_test()


def test_plotnine_all_imports():
    import plotnine as p9

    for name in p9.__all__:
        m = getattr(p9, name).__module__
        assert m.startswith("plotnine"), f"{m} in plotnine.__all__!"



from __future__ import annotations

import typing
from copy import copy

import numpy as np

from ..iapi import panel_ranges

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt
    import pandas as pd

    from plotnine.iapi import labels_view, panel_view
    from plotnine.typing import (
        FloatArray,
        FloatArrayLike,
        Ggplot,
        Scale,
    )


class coord:
    """
    Base class for all coordinate systems
    """

    # If the coordinate system is linear
    is_linear = False

    # Additional parameters created acc. to the data,
    # if the coordinate system needs them
    params: dict[str, Any]

    def __radd__(self, gg: Ggplot) -> Ggplot:
        """
        Add coordinates to ggplot object
        """
        gg.coordinates = copy(self)
        return gg

    def setup_data(self, data: list[pd.DataFrame]) -> list[pd.DataFrame]:
        """
        Allow the coordinate system to manipulate the layer data

        Parameters
        ----------
        data : list[DataFrames]
            Data for alls Layer

        Returns
        -------
        out : list[DataFrame]
            Modified layer data
        """
        return data

    def setup_params(self, data: list[pd.DataFrame]):
        """
        Create additional parameters

        A coordinate system may need to create parameters
        depending on the *original* data that the layers get.

        Parameters
        ----------
        data : list of dataframes
            Data for each layer before it is manipulated in
            any way.
        """
        self.params = {}

    def setup_layout(self, layout: pd.DataFrame) -> pd.DataFrame:
        """
        Allow the coordinate system alter the layout dataframe

        Parameters
        ----------
        layout : dataframe
            Dataframe in which data is assigned to panels and scales

        Returns
        -------
        out : dataframe
            layout dataframe altered to according to the requirements
            of the coordinate system.

        Notes
        -----
        The input dataframe may be changed.
        """
        return layout

    def aspect(self, panel_params: panel_view) -> float | None:
        """
        Return desired aspect ratio for the plot

        If not overridden by the subclass, this method
        returns ``None``, which means that the coordinate
        system does not influence the aspect ratio.
        """
        return None

    def labels(self, cur_labels: labels_view) -> labels_view:
        """
        Modify labels

        Parameters
        ----------
        cur_labels : labels_view
            Current labels. The coord can modify them as necessary.

        Returns
        -------
        out : dict
            Modified labels. Same object as the input.
        """
        return cur_labels

    def transform(
        self, data: pd.DataFrame, panel_params: panel_view, munch: bool = False
    ) -> pd.DataFrame:
        """
        Transform data before it is plotted

        This is used to "transform the coordinate axes".
        Subclasses should override this method
        """
        return data

    def setup_panel_params(self, scale_x: Scale, scale_y: Scale) -> panel_view:
        """
        Compute the range and break information for the panel
        """
        msg = "The coordinate should implement this method."
        raise NotImplementedError(msg)

    def range(self, panel_params: panel_view) -> panel_ranges:
        """
        Return the range along the dimensions of the coordinate system
        """
        # Defaults to providing the 2D x-y ranges
        return panel_ranges(x=panel_params.x.range, y=panel_params.y.range)

    def backtransform_range(self, panel_params: panel_view) -> panel_ranges:
        """
        Backtransform the panel range in panel_params to data coordinates

        Coordinate systems that do any transformations should override
        this method. e.g. coord_trans has to override this method.
        """
        return self.range(panel_params)

    def distance(
        self,
        x: pd.Series[float],
        y: pd.Series[float],
        panel_params: panel_view,
    ) -> npt.NDArray[Any]:
        msg = "The coordinate should implement this method."
        raise NotImplementedError(msg)

    def munch(
        self, data: pd.DataFrame, panel_params: panel_view
    ) -> pd.DataFrame:
        ranges = self.backtransform_range(panel_params)

        data.loc[data["x"] == -np.inf, "x"] = ranges.x[0]
        data.loc[data["x"] == np.inf, "x"] = ranges.x[1]
        data.loc[data["y"] == -np.inf, "y"] = ranges.y[0]
        data.loc[data["y"] == np.inf, "y"] = ranges.y[1]

        dist = self.distance(data["x"], data["y"], panel_params)
        bool_idx = (
            data["group"].iloc[1:].values != data["group"].iloc[:-1].values
        )
        dist[bool_idx] = np.nan

        # Munch
        munched = munch_data(data, dist)
        return munched


def dist_euclidean(x: FloatArrayLike, y: FloatArrayLike) -> FloatArray:
    """
    Calculate euclidean distance
    """
    x = np.asarray(x, dtype=np.float64)
    y = np.asarray(y, dtype=np.float64)
    return np.sqrt(
        (x[:-1] - x[1:]) ** 2 + (y[:-1] - y[1:]) ** 2, dtype=np.float64
    )


def interp(start: int, end: int, n: int) -> FloatArray:
    """
    Interpolate
    """
    return np.linspace(start, end, n, endpoint=False)


def munch_data(data: pd.DataFrame, dist: FloatArray) -> pd.DataFrame:
    """
    Breakup path into small segments
    """
    x, y = data["x"], data["y"]
    segment_length = 0.01

    # How many endpoints for each old segment,
    # not counting the last one
    dist[np.isnan(dist)] = 1
    extra = np.maximum(np.floor(dist / segment_length), 1)
    extra = extra.astype(int, copy=False)

    # Generate extra pieces for x and y values
    # The final point must be manually inserted at the end
    x = [interp(start, end, n) for start, end, n in zip(x[:-1], x[1:], extra)]
    y = [interp(start, end, n) for start, end, n in zip(y[:-1], y[1:], extra)]
    x.append(data["x"].iloc[-1])
    y.append(data["y"].iloc[-1])
    x = np.hstack(x)
    y = np.hstack(y)

    # Replicate other aesthetics: defined by start point
    # but also must include final point
    idx = np.hstack(
        [
            np.repeat(data.index[:-1], extra),
            len(data) - 1
            # data.index[-1] # TODO: Maybe not
        ]
    )

    munched = data.loc[idx, list(data.columns.difference(["x", "y"]))]
    munched["x"] = x
    munched["y"] = y
    munched.reset_index(drop=True, inplace=True)

    return munched



from __future__ import annotations

import typing
from itertools import chain

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..iapi import range_view
from ..utils import alias, array_kind, match
from ._expand import expand_range
from .range import RangeContinuous
from .scale import scale_continuous, scale_datetime, scale_discrete

if typing.TYPE_CHECKING:
    from typing import Sequence

    from plotnine.typing import (
        Trans,
        TupleFloat2,
        TupleFloat4,
    )


# positions scales have a couple of differences (quirks) that
# make necessary to override some of the scale_discrete and
# scale_continuous methods
#
# scale_position_discrete and scale_position_continuous
# are intermediate base classes where the required overriding
# is done
@document
class scale_position_discrete(scale_discrete):
    """
    Base class for discrete position scales

    Parameters
    ----------
    {superclass_parameters}
    limits : array_like, optional
        Limits of the scale. For discrete scale, these are
        the categories (unique values) of the variable.
        For scales that deal with categoricals, these may
        be a subset or superset of the categories.
    """

    # All positions have no guide
    guide = None

    # Keeps two ranges, range and range_c
    range_c: RangeContinuous

    def __init__(self, *args, **kwargs):
        self.range_c = RangeContinuous()
        scale_discrete.__init__(self, *args, **kwargs)

    def reset(self):
        # Can't reset discrete scale because
        # no way to recover values
        self.range_c.reset()

    def is_empty(self) -> bool:
        return super().is_empty() and self.range_c.is_empty()

    def train(self, series):
        # The discrete position scale is capable of doing
        # training for continuous data.
        # This complicates training and mapping, but makes it
        # possible to place objects at non-integer positions,
        # as is necessary for jittering etc.
        if array_kind.continuous(series):
            self.range_c.train(series)
        else:
            self.range.train(series, drop=self.drop)

    def map(self, series, limits=None):
        # Discrete values are converted into integers starting
        # at 1
        if limits is None:
            limits = self.limits
        if array_kind.discrete(series):
            # TODO: Rewrite without using numpy
            seq = np.arange(1, len(limits) + 1)
            idx = np.asarray(match(series, limits, nomatch=len(series)))
            if not len(idx):
                return np.array([])
            try:
                seq = seq[idx]
            except IndexError:
                # Deal with missing data
                # - Insert NaN where there is no match
                seq = np.hstack((seq.astype(float), np.nan))
                idx = np.clip(idx, 0, len(seq) - 1)
                seq = seq[idx]
            return seq
        return series

    @property
    def limits(self):
        if self.is_empty():
            return (0, 1)
        elif self._limits is not None and not callable(self._limits):
            return self._limits
        elif self._limits is None:
            # discrete range
            return self.range.range
        elif callable(self._limits):
            limits = self._limits(self.range.range)
            # Functions that return iterators e.g. reversed
            if iter(limits) is limits:
                limits = list(limits)
            return limits
        else:
            raise PlotnineError("Lost, do not know what the limits are.")

    @limits.setter
    def limits(self, value):
        if isinstance(value, tuple):
            value = list(value)
        self._limits = value

    def dimension(self, expand=(0, 0, 0, 0), limits=None):
        """
        Get the phyical size of the scale

        Unlike limits, this always returns a numeric vector of length 2
        """
        from mizani.bounds import expand_range_distinct

        if limits is None:
            limits = self.limits

        if self.is_empty():
            return (0, 1)

        if self.range.is_empty():  # only continuous
            return expand_range_distinct(self.range_c.range, expand)
        elif self.range_c.is_empty():  # only discrete
            # FIXME: I think this branch should not exist
            return expand_range_distinct((1, len(self.limits)), expand)
        else:  # both
            # e.g categorical bar plot have discrete items, but
            # are plot on a continuous x scale
            a = np.hstack(
                [
                    self.range_c.range,
                    expand_range_distinct((1, len(self.range.range)), expand),
                ]
            )
            return a.min(), a.max()

    def expand_limits(
        self,
        limits: Sequence[str],
        expand: TupleFloat2 | TupleFloat4,
        coord_limits: TupleFloat2,
        trans: Trans,
    ) -> range_view:
        # Turn discrete limits into a tuple of continuous limits
        if self.is_empty():
            climits = (0, 1)
        else:
            climits = (1, len(limits))
            self.range_c.range

        if coord_limits is not None:
            # - Override None in coord_limits
            # - Expand limits in coordinate space
            # - Remove any computed infinite values &
            c0, c1 = coord_limits
            climits = (
                climits[0] if c0 is None else c0,
                climits[1] if c1 is None else c1,
            )

        # Expand discrete range
        rv_d = expand_range(climits, expand, trans)

        if self.range_c.is_empty():
            return rv_d

        # Expand continuous range
        no_expand = self.default_expansion(0, 0)
        rv_c = expand_range(self.range_c.range, no_expand, trans)

        # Merge the ranges
        rv = range_view(
            range=(
                min(chain(rv_d.range, rv_c.range)),
                max(chain(rv_d.range, rv_c.range)),
            ),
            range_coord=(
                min(chain(rv_d.range_coord, rv_c.range_coord)),
                max(chain(rv_d.range_coord, rv_c.range_coord)),
            ),
        )
        rv.range = min(rv.range), max(rv.range)
        rv.range_coord = min(rv.range_coord), max(rv.range_coord)
        return rv


@document
class scale_position_continuous(scale_continuous):
    """
    Base class for continuous position scales

    Parameters
    ----------
    {superclass_parameters}
    """

    # All positions have no guide
    guide = None

    def map(self, series, limits=None):
        # Position aesthetics don't map, because the coordinate
        # system takes care of it.
        # But the continuous scale has to deal with out of bound points
        if not len(series):
            return series
        if limits is None:
            limits = self.limits
        scaled = self.oob(series, limits)
        scaled[pd.isnull(scaled)] = self.na_value
        return scaled


@document
class scale_x_discrete(scale_position_discrete):
    """
    Discrete x position

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["x", "xmin", "xmax", "xend"]


@document
class scale_y_discrete(scale_position_discrete):
    """
    Discrete y position

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["y", "ymin", "ymax", "yend"]


# Not part of the user API
alias("scale_x_ordinal", scale_x_discrete)
alias("scale_y_ordinal", scale_y_discrete)


@document
class scale_x_continuous(scale_position_continuous):
    """
    Continuous x position

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["x", "xmin", "xmax", "xend", "xintercept"]


@document
class scale_y_continuous(scale_position_continuous):
    """
    Continuous y position

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = [
        "y",
        "ymin",
        "ymax",
        "yend",
        "yintercept",
        "ymin_final",
        "ymax_final",
        "lower",
        "middle",
        "upper",
    ]  # pyright: ignore


# Transformed scales
@document
class scale_x_datetime(scale_datetime, scale_x_continuous):
    """
    Continuous x position for datetime data points

    Parameters
    ----------
    {superclass_parameters}
    """


@document
class scale_y_datetime(scale_datetime, scale_y_continuous):
    """
    Continuous y position for datetime data points

    Parameters
    ----------
    {superclass_parameters}
    """


alias("scale_x_date", scale_x_datetime)
alias("scale_y_date", scale_y_datetime)


@document
class scale_x_timedelta(scale_x_continuous):
    """
    Continuous x position for timedelta data points

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "pd_timedelta"


@document
class scale_y_timedelta(scale_y_continuous):
    """
    Continuous y position for timedelta data points

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "pd_timedelta"


@document
class scale_x_sqrt(scale_x_continuous):
    """
    Continuous x position sqrt transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "sqrt"


@document
class scale_y_sqrt(scale_y_continuous):
    """
    Continuous y position sqrt transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "sqrt"


@document
class scale_x_log10(scale_x_continuous):
    """
    Continuous x position log10 transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "log10"


@document
class scale_y_log10(scale_y_continuous):
    """
    Continuous y position log10 transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "log10"


@document
class scale_x_reverse(scale_x_continuous):
    """
    Continuous x position reverse transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "reverse"


@document
class scale_y_reverse(scale_y_continuous):
    """
    Continuous y position reverse transformed scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _trans = "reverse"



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..utils import copy_missing_columns, resolution
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_errorbar(geom):
    """
    Vertical interval represented as an errorbar

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float or None, optional (default: 0.5)
        Bar width. If :py:`None`, the width is set to
        `90%` of the resolution of the data.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
    }
    REQUIRED_AES = {"x", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "width": 0.5,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "width" not in data:
            if self.params["width"]:
                data["width"] = self.params["width"]
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        data["xmin"] = data["x"] - data["width"] / 2
        data["xmax"] = data["x"] + data["width"] / 2
        del data["width"]
        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        f = np.hstack
        # create (two horizontal bars) + vertical bar
        df = pd.DataFrame(
            {
                "x": f([data["xmin"], data["xmin"], data["x"]]),
                "xend": f([data["xmax"], data["xmax"], data["x"]]),
                "y": f([data["ymin"], data["ymax"], data["ymax"]]),
                "yend": f([data["ymin"], data["ymax"], data["ymin"]]),
            }
        )

        copy_missing_columns(df, data)
        geom_segment.draw_group(df, panel_params, coord, ax, **params)



from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod

from ..exceptions import PlotnineError
from ..iapi import strip_label_details

if typing.TYPE_CHECKING:
    from typing import Callable, Optional

    from ..typing import (
        CanBeStripLabellingFunc,
        StripLabellingFunc,
        StripLabellingFuncNames,
    )


def label_value(
    label_info: strip_label_details, multi_line: bool = True
) -> strip_label_details:
    """
    Keep value as the label

    Parameters
    ----------
    label_info : strip_label_details
        Label information whose values will be returned
    multi_line : bool
        Whether to place each variable on a separate line

    Returns
    -------
    out : strip_label_details
        Label text strings
    """
    label_info = label_info.copy()

    if not multi_line:
        label_info = label_info.collapse()

    return label_info


def label_both(
    label_info: strip_label_details, multi_line: bool = True, sep: str = ": "
) -> strip_label_details:
    """
    Concatenate the facet variable with the value

    Parameters
    ----------
    label_info : strip_label_details
        Label information to be modified.
    multi_line : bool
        Whether to place each variable on a separate line
    sep :  str
        Separation between variable name and value

    Returns
    -------
    out : strip_label_details
        Label information
    """
    label_info = label_info.copy()

    for var, lvalue in label_info.variables.items():
        label_info.variables[var] = f"{var}{sep}{lvalue}"

    if not multi_line:
        label_info = label_info.collapse()

    return label_info


def label_context(
    label_info: strip_label_details, multi_line: bool = True, sep: str = ": "
) -> strip_label_details:
    """
    Create an unabiguous label string

    If facetting over a single variable, `label_value` is
    used, if two or more variables then `label_both` is used.

    Parameters
    ----------
    label_info : strip_label_details
        Label information
    multi_line : bool
        Whether to place each variable on a separate line
    sep :  str
        Separation between variable name and value

    Returns
    -------
    out : str
        Contatenated label values (or pairs of variable names
        & values)
    """
    if len(label_info) == 1:
        return label_value(label_info, multi_line)
    else:
        return label_both(label_info, multi_line, sep)


LABELLERS: dict[StripLabellingFuncNames, StripLabellingFunc] = {
    "label_value": label_value,
    "label_both": label_both,
    "label_context": label_context,
}


def as_labeller(
    x: Optional[CanBeStripLabellingFunc] = None,
    default: CanBeStripLabellingFunc = label_value,
    multi_line: bool = True,
) -> labeller:
    """
    Coerse to labeller

    Parameters
    ----------
    x : function | dict
        Object to coerce
    default : str | function
        Default labeller. If it is a string,
        it should be the name of one the labelling
        functions provided by plotnine.
    multi_line : bool
        Whether to place each variable on a separate line

    Returns
    -------
    out : labeller
        Labelling function
    """
    if x is None:
        x = default

    if isinstance(x, labeller):
        return x

    x = _as_strip_labelling_func(x)
    return labeller(rows=x, cols=x, multi_line=multi_line)


class labeller:
    """
    Facet Strip Labelling

    When called with strip_label_details knows how to
    alter the strip labels along either dimension.

    Parameters
    ----------
    rows : str | function | None
        How to label the rows
    cols : str | function | None
        How to label the columns
    multi_line : bool
        Whether to place each variable on a separate line
    default : function | str
        Fallback labelling function. If it is a string,
        it should be the name of one the labelling
        functions provided by plotnine.
    kwargs : dict
        {variable name : function | string} pairs for
        renaming variables. A function to rename the variable
        or a string name.
    """

    def __init__(
        self,
        rows: Optional[CanBeStripLabellingFunc] = None,
        cols: Optional[CanBeStripLabellingFunc] = None,
        multi_line: bool = True,
        default: CanBeStripLabellingFunc = "label_value",
        **kwargs: Callable[[str], str],
    ):
        # Sort out the labellers along each dimension
        self.rows_labeller = _as_strip_labelling_func(rows, default)
        self.cols_labeller = _as_strip_labelling_func(cols, default)
        self.multi_line = multi_line
        self.variable_maps = kwargs

    def __call__(self, label_info: strip_label_details) -> strip_label_details:
        """
        Called to do the labelling
        """
        variable_maps = {
            k: v
            for k, v in self.variable_maps.items()
            if k in label_info.variables
        }

        # No variable specific labeller
        if label_info.meta["dimension"] == "rows":
            result = self.rows_labeller(label_info)
        else:
            result = self.cols_labeller(label_info)

        # Make dict_labeler for the  variable specific labelers
        # do the label and merge
        if variable_maps:
            d = {
                value: variable_maps[var]
                for var, value in label_info.variables.items()
                if var in variable_maps
            }
            func = _as_strip_labelling_func(d)
            result2 = func(label_info)
            result.variables.update(result2.variables)

        if not self.multi_line:
            result = result.collapse()

        return result


def _as_strip_labelling_func(
    fobj: Optional[CanBeStripLabellingFunc],
    default: CanBeStripLabellingFunc = "label_value",
) -> StripLabellingFunc:
    """
    Create a function that can operate on strip_label_details
    """
    if fobj is None:
        fobj = default

    if isinstance(fobj, str) and fobj in LABELLERS:
        return LABELLERS[fobj]

    if isinstance(fobj, _core_labeller):
        return fobj
    elif callable(fobj):
        if fobj.__name__ in LABELLERS:
            return fobj  # type: ignore
        else:
            return _function_labeller(fobj)  # type: ignore
    elif isinstance(fobj, dict):
        return _dict_labeller(fobj)
    else:
        msg = f"Could not create a labelling function for with `{fobj}`."
        raise PlotnineError(msg)


class _core_labeller(metaclass=ABCMeta):
    """
    Per item
    """

    @abstractmethod
    def __call__(self, label_info: strip_label_details) -> strip_label_details:
        pass


class _function_labeller(_core_labeller):
    """
    Use a function turn facet value into a label

    Parameters
    ----------
    func : callable
        Function to label an individual string
    """

    def __init__(self, func: Callable[[str], str]):
        self.func = func

    def __call__(self, label_info: strip_label_details) -> strip_label_details:
        label_info = label_info.copy()
        variables = label_info.variables
        for facet_var, facet_value in variables.items():
            variables[facet_var] = self.func(facet_value)
        return label_info


class _dict_labeller(_core_labeller):
    """
    Use a dict to alter specific facet values

    Parameters
    ----------
    lookup : dict
        A dict of the one of the forms
          - {facet_value: label_value}
          - {facet_value: callable(<label_value>)}
    """

    def __init__(
        self, lookup: dict[str, str] | dict[str, Callable[[str], str]]
    ):
        self.lookup = lookup

    def __call__(self, label_info: strip_label_details) -> strip_label_details:
        label_info = label_info.copy()
        variables = label_info.variables
        # Replace facet_value with values from the lookup table
        # If the value is function, call it  the result of calling function
        for facet_var, facet_value in variables.items():
            target = self.lookup.get(facet_value)
            if target is None:
                continue
            elif callable(target):
                variables[facet_var] = target(facet_value)
            else:
                variables[facet_var] = target
        return label_info



import pandas as pd

from plotnine import aes, geom_line, ggplot, scale_linetype_manual


def test_scale_linetype_manual_tuples():
    # linetype_manual accepts tuples as mapping results
    # this must be tested specifically.
    df = pd.DataFrame(
        {
            "x": [0, 1, 0, 1, 0, 1],
            "y": [0, 1, 0, 2, 0, 3],
            "lt": ["A", "A", "B", "B", "C", "C"],
        }
    )

    p = ggplot(df)
    p += geom_line(aes(x="x", y="y", ymax="y+1", linetype="lt", group="lt"))
    p += scale_linetype_manual(
        values=(
            (2, (5, 3, 1, 3)),
            (1, (10, 2)),
            (1, (1, 2)),
        )
    )
    assert p == "scale_linetype_manual_tuples"


def test_scale_linetype_strings_tuples():
    # linetype_manual accepts tuples as mapping results
    # this must be tested specifically.
    df = pd.DataFrame(
        {
            "x": [0, 1, 0, 1, 0, 1],
            "y": [0, 1, 0, 2, 0, 3],
            "lt": ["A", "A", "B", "B", "C", "C"],
        }
    )

    p = ggplot(df)
    p += geom_line(aes(x="x", y="y", ymax="y+1", linetype="lt", group="lt"))
    p += scale_linetype_manual(values=["solid", "dashed", "dotted"])
    assert p == "scale_linetype_manual_strings"



import inspect
import locale
import shutil
import types
import warnings
from copy import deepcopy
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.testing.compare import compare_images

from plotnine import ggplot, theme

TOLERANCE = 2  # Default tolerance for the tests
DPI = 72  # Default DPI for the tests

# This partial theme modifies all themes that are used in
# the test. It is limited to setting the size of the test
# images Should a test require a larger or smaller figure
# size, the dpi or aspect_ratio should be modified.
test_theme = theme(figure_size=(640 / DPI, 480 / DPI), dpi=DPI)

tests_dir = Path(__file__).parent
baseline_images_dir = tests_dir / "baseline_images"
result_images_dir = tests_dir / "result_images"

if not baseline_images_dir.exists():
    raise OSError(
        "The baseline image directory does not exist. "
        "This is most likely because the test data is not installed. "
        "You may need to install plotnine from source to get the "
        "test data."
    )


def raise_no_baseline_image(filename: str):
    raise Exception(f"Baseline image {filename} is missing")


def ggplot_equals(gg: ggplot, name: str) -> bool:
    """
    Compare ggplot object to image determined by `right`

    Parameters
    ----------
    gg : ggplot
        ggplot object
    name : str
        Identifier for the test image

    This function is meant to monkey patch ggplot.__eq__
    so that tests can use the `assert` statement.
    """
    test_file = inspect.stack()[1][1]
    filenames = make_test_image_filenames(name, test_file)
    # Save the figure before testing whether the original image
    # actually exists. This makes creating new tests much easier,
    # as the result image can afterwards just be copied.
    gg += test_theme
    with _test_cleanup():
        gg.save(filenames.result, verbose=False)

    if filenames.baseline.exists():
        shutil.copyfile(filenames.baseline, filenames.expected)
    else:
        # Putting the exception in short function makes for
        #  short pytest error messages
        raise_no_baseline_image(filenames.baseline)

    err = compare_images(
        filenames.expected, filenames.result, TOLERANCE, in_decorator=True
    )
    gg._err = err  # For the pytest error message
    return False if err else True


ggplot.__eq__ = ggplot_equals


def draw_test(self):
    """
    Try drawing the ggplot object

    Parameters
    ----------
    self : ggplot
        ggplot object

    This function is meant to monkey patch ggplot.draw_test
    so that tests can draw and not care about cleaning up
    the MPL figure.
    """
    with _test_cleanup():
        self.draw()


ggplot.draw_test = draw_test


def build_test(self):
    """
    Try building the ggplot object

    Parameters
    ----------
    self : ggplot
        ggplot object

    This function is meant to monkey patch ggplot.build_test
    so that tests can build a plot and inspect the side effects
    on the plot object.
    """
    self = deepcopy(self)
    self._build()
    return self


ggplot.build_test = build_test


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, ggplot) and isinstance(right, str) and op == "==":
        msg = (
            "images not close: {actual:s} vs. {expected:s} "
            "(RMS {rms:.2f})".format(**left._err)
        )
        return [msg]


def make_test_image_filenames(name, test_file):
    """
    Create filenames for testing

    Parameters
    ----------
    name : str
        An identifier for the specific test. This will make-up
        part of the filenames.
    test_file : str
        Full path of the test file. This will determine the
        directory structure

    Returns
    -------
    out : types.SimpleNamespace
        Object with 3 attributes to store the generated filenames

            - result
            - baseline
            - expected

        `result`, is the filename for the image generated by the test.
        `baseline`, is the filename for the baseline image to which
        the result will be compared.
        `expected`, is the filename to the copy of the baseline that
        will be stored in the same directory as the result image.
        Creating a copy make comparison easier.
    """
    name = Path(name).with_suffix(".png")
    expected_name = f"{name.stem}-expected{name.suffix}"
    subdir = Path(test_file).stem
    filenames = types.SimpleNamespace(
        baseline=baseline_images_dir / subdir / name,
        result=result_images_dir / subdir / name,
        expected=result_images_dir / subdir / expected_name,
    )
    filenames.result.parent.mkdir(parents=True, exist_ok=True)
    return filenames


class _test_cleanup:
    def __enter__(self):
        # The baseline images are created in this locale, so we should use
        # it during all of the tests.
        try:
            locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
        except locale.Error:
            try:
                locale.setlocale(locale.LC_ALL, "English_United States.1252")
            except locale.Error:
                warnings.warn(
                    "Could not set locale to English/United States. "
                    "Some date-related tests may fail"
                )

        # make sure we don't carry over bad plots from former tests
        plt.close("all")
        n_figs = len(plt.get_fignums())
        msg = (
            f"No. of open figs: {n_figs}. Make sure the "
            "figures from the previous tests are cleaned up."
        )
        assert n_figs == 0, msg

        mpl.use("Agg")
        # These settings *must* be hardcoded for running the comparison
        # tests
        mpl.rcdefaults()  # Start with all defaults
        mpl.rcParams["text.hinting"] = "auto"
        mpl.rcParams["text.antialiased"] = True
        mpl.rcParams["text.hinting_factor"] = 8
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        plt.close("all")
        warnings.resetwarnings()


def layer_data(p, i=0):
    """
    Return layer information used to draw the plot

    Parameters
    ----------
    p : ggplot
        ggplot object
    i : int
        Layer number

    Returns
    -------
    out : dataframe
        Layer information
    """
    p = deepcopy(p)
    p._build()
    return p.layers.data[i]



from __future__ import annotations

import typing
import warnings

import numpy as np
import pandas as pd

from ..coords import coord_flip
from ..exceptions import PlotnineWarning
from ..scales.scale import scale_continuous as ScaleContinuous
from ..utils import log
from .annotate import annotate
from .geom_path import geom_path
from .geom_rug import geom_rug

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional, Sequence

    from typing_extensions import TypeGuard

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        AnyArray,
        Axes,
        Coord,
        Geom,
        Layout,
        Scale,
        Trans,
        TupleFloat2,
        TupleFloat3,
    )


class _geom_logticks(geom_rug):
    """
    Internal geom implementing drawing of annotation_logticks
    """

    DEFAULT_AES = {}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "sides": "bl",
        "alpha": 1,
        "color": "black",
        "size": 0.5,
        "linetype": "solid",
        "lengths": (0.036, 0.0225, 0.012),
        "base": 10,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    def draw_layer(
        self, data: pd.DataFrame, layout: Layout, coord: Coord, **params: Any
    ):
        """
        Draw ticks on every panel
        """
        for pid in layout.layout["PANEL"]:
            ploc = pid - 1
            panel_params = layout.panel_params[ploc]
            ax = layout.axs[ploc]
            self.draw_panel(data, panel_params, coord, ax, **params)

    @staticmethod
    def _check_log_scale(
        base: Optional[float],
        sides: str,
        panel_params: panel_view,
        coord: Coord,
    ) -> TupleFloat2:
        """
        Check the log transforms

        Parameters
        ----------
        base : float or None
            Base of the logarithm in which the ticks will be
            calculated. If ``None``, the base of the log transform
            the scale will be used.
        sides : str (default: bl)
            Sides onto which to draw the marks. Any combination
            chosen from the characters ``btlr``, for *bottom*, *top*,
            *left* or *right* side marks. If ``coord_flip()`` is used,
            these are the sides *before* the flip.
        panel_params : panel_view
            ``x`` and ``y`` view scale values.
        coord : coord
            Coordinate (e.g. coord_cartesian) system of the geom.

        Returns
        -------
        out : tuple
            The bases (base_x, base_y) to use when generating the ticks.
        """

        def is_log_trans(t: Trans) -> bool:
            return hasattr(t, "base") and t.__class__.__name__.startswith(
                "log"
            )

        def get_base(sc, ubase: Optional[float]) -> float:
            ae = sc.aesthetics[0]

            if not isinstance(sc, ScaleContinuous) or not is_log_trans(
                sc.trans
            ):
                warnings.warn(
                    f"annotation_logticks for {ae}-axis which does not have "
                    "a log scale. The logticks may not make sense.",
                    PlotnineWarning,
                )
                return 10 if ubase is None else ubase

            base = sc.trans.base  # pyright: ignore
            if ubase is not None and base != ubase:
                warnings.warn(
                    f"The x-axis is log transformed in base={base} ,"
                    "but the annotation_logticks are computed in base="
                    f"{ubase}",
                    PlotnineWarning,
                )
                return ubase
            return base

        base_x, base_y = 10, 10
        x_scale = panel_params.x.scale
        y_scale = panel_params.y.scale

        if isinstance(coord, coord_flip):
            x_scale, y_scale = y_scale, x_scale
            base_x, base_y = base_y, base_x

        if "t" in sides or "b" in sides:
            base_x = get_base(x_scale, base)

        if "l" in sides or "r" in sides:
            base_y = get_base(y_scale, base)

        return base_x, base_y

    @staticmethod
    def _calc_ticks(
        value_range: TupleFloat2, base: float
    ) -> tuple[AnyArray, AnyArray, AnyArray]:
        """
        Calculate tick marks within a range

        Parameters
        ----------
        value_range: tuple
            Range for which to calculate ticks.

        base : number
            Base of logarithm

        Returns
        -------
        out: tuple
            (major, middle, minor) tick locations
        """

        def _minor(x: Sequence[Any], mid_idx: int) -> AnyArray:
            return np.hstack([x[1:mid_idx], x[mid_idx + 1 : -1]])

        # * Calculate the low and high powers,
        # * Generate for all intervals in along the low-high power range
        #   The intervals are in normal space
        # * Calculate evenly spaced breaks in normal space, then convert
        #   them to log space.
        low = np.floor(value_range[0])
        high = np.ceil(value_range[1])
        arr = base ** np.arange(low, float(high + 1))
        n_ticks = int(np.round(base) - 1)
        breaks = [
            log(np.linspace(b1, b2, n_ticks + 1), base)
            for (b1, b2) in list(zip(arr, arr[1:]))
        ]

        # Partition the breaks in the 3 groups
        major = np.array([x[0] for x in breaks] + [breaks[-1][-1]])
        if n_ticks % 2:
            mid_idx = n_ticks // 2
            middle = np.array([x[mid_idx] for x in breaks])
            minor = np.hstack([_minor(x, mid_idx) for x in breaks])
        else:
            middle = np.array([])
            minor = np.hstack([x[1:-1] for x in breaks])

        return major, middle, minor

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        # Any passed data is ignored, the relevant data is created
        sides = params["sides"]
        lengths = params["lengths"]
        _aesthetics = {
            "size": params["size"],
            "color": params["color"],
            "alpha": params["alpha"],
            "linetype": params["linetype"],
        }

        def _draw(
            geom: Geom,
            axis: Literal["x", "y"],
            tick_positions: tuple[AnyArray, AnyArray, AnyArray],
        ):
            for position, length in zip(tick_positions, lengths):
                data = pd.DataFrame({axis: position, **_aesthetics})
                geom.draw_group(
                    data, panel_params, coord, ax, length=length, **params
                )

        if isinstance(coord, coord_flip):
            tick_range_x = panel_params.y.range
            tick_range_y = panel_params.x.range
        else:
            tick_range_x = panel_params.x.range
            tick_range_y = panel_params.y.range

        # these are already flipped iff coord_flip
        base_x, base_y = self._check_log_scale(
            params["base"], sides, panel_params, coord
        )

        if "b" in sides or "t" in sides:
            tick_positions = self._calc_ticks(tick_range_x, base_x)
            _draw(self, "x", tick_positions)

        if "l" in sides or "r" in sides:
            tick_positions = self._calc_ticks(tick_range_y, base_y)
            _draw(self, "y", tick_positions)


class annotation_logticks(annotate):
    """
    Marginal log ticks.

    If added to a plot that does not have a log10 axis
    on the respective side, a warning will be issued.

    Parameters
    ----------
    sides : str (default: bl)
        Sides onto which to draw the marks. Any combination
        chosen from the characters ``btlr``, for *bottom*, *top*,
        *left* or *right* side marks. If ``coord_flip()`` is used,
        these are the sides *after* the flip.
    alpha : float (default: 1)
        Transparency of the ticks
    color : str | tuple (default: 'black')
        Colour of the ticks
    size : float
        Thickness of the ticks
    linetype : 'solid' | 'dashed' | 'dashdot' | 'dotted' | tuple
        Type of line. Default is *solid*.
    lengths: tuple (default (0.036, 0.0225, 0.012))
        length of the ticks drawn for full / half / tenth
        ticks relative to panel size
    base : float (default: None)
        Base of the logarithm in which the ticks will be
        calculated. If ``None``, the base used to log transform
        the scale will be used.
    """

    def __init__(
        self,
        sides: str = "bl",
        alpha: float = 1,
        color: str | tuple[float, ...] = "black",
        size: float = 0.5,
        linetype: str | tuple[float, ...] = "solid",
        lengths: TupleFloat3 = (0.036, 0.0225, 0.012),
        base: float | None = None,
    ):
        if len(lengths) != 3:
            raise ValueError(
                "length for annotation_logticks must be a tuple of 3 floats"
            )

        self._annotation_geom = _geom_logticks(
            sides=sides,
            alpha=alpha,
            color=color,
            size=size,
            linetype=linetype,
            lengths=lengths,
            base=base,
        )


def is_continuous_scale(sc: Scale) -> TypeGuard[ScaleContinuous]:
    return isinstance(sc, ScaleContinuous)



from __future__ import annotations

import typing
from warnings import warn

import numpy as np

from ..coords import coord_cartesian
from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import resolution
from .geom import geom
from .geom_polygon import geom_polygon

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Aes, Axes, Coord, DataLike


@document
class geom_raster(geom):
    """
    Rasterized Rectangles specified using center points

    {usage}

    Parameters
    ----------
    {common_parameters}

    hjust : float (default: 0.5)
        Horizontal justification for the rectangle at point ``x``.
        Default is 0.5, which centers the rectangle horizontally.
        Must be in the range ``[0, 1]``.
    vjust : float (default: 0.5)
        Vertical justification for the rectangle at point ``y``
        Default is 0.5, which centers the rectangle vertically.
        Must be in the range ``[0, 1]``.
    interpolation : str | None (default: None)
        How to calculate values between the center points of
        adjacent rectangles. The default is :py:`None` not to
        interpolate. Allowed values are:
        :py:`'none', 'antialiased', 'nearest', 'bilinear', 'bicubic',`
        :py:`'spline16', 'spline36', 'hanning', 'hamming', 'hermite',`
        :py:`'kaiser', 'quadric', 'catrom', 'gaussian', 'bessel',`
        :py:`'mitchell', 'sinc', 'lanczos', 'blackman'`

    filterrad : float, (default: 4.0)
        The filter radius for filters that have a radius parameter, i.e.
        when interpolation is one of: *'sinc', 'lanczos' or 'blackman'*.
        Must be a number greater than zero.

    See Also
    --------
    plotnine.geoms.geom_rect
    plotnine.geoms.geom_tile
    """

    DEFAULT_AES = {"alpha": 1, "fill": "#333333"}
    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"fill", "xmin", "xmax", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "vjust": 0.5,
        "hjust": 0.5,
        "interpolation": None,
        "filterrad": 4.0,
        "raster": True,
    }
    draw_legend = staticmethod(geom_polygon.draw_legend)

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        # Silently accept:
        #    1. interpolate
        #    2. bool values for interpolation
        if "interpolate" in kwargs:
            kwargs["interpolation"] = kwargs.pop("interpolate")
        if isinstance(kwargs.get("interpolation", None), bool):
            if kwargs["interpolation"] is True:
                kwargs["interpolation"] = "bilinear"
            else:
                kwargs["interpolation"] = None

        super().__init__(mapping, data, **kwargs)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        hjust = self.params["hjust"]
        vjust = self.params["vjust"]
        precision = np.sqrt(np.finfo(float).eps)

        x_diff = np.diff(np.sort(data["x"].unique()))
        if len(x_diff) == 0:
            w = 1
        elif np.any(np.abs(np.diff(x_diff)) > precision):
            warn(
                "Raster pixels are placed at uneven horizontal intervals "
                "and will be shifted. Consider using geom_tile() instead.",
                PlotnineWarning,
            )
            w = x_diff.min()
        else:
            w = x_diff[0]

        y_diff = np.diff(np.sort(data["y"].unique()))
        if len(y_diff) == 0:
            h = 1
        elif np.any(np.abs(np.diff(y_diff)) > precision):
            warn(
                "Raster pixels are placed at uneven vertical intervals "
                "and will be shifted. Consider using geom_tile() instead.",
                PlotnineWarning,
            )
            h = y_diff.min()
        else:
            h = y_diff[0]

        data["xmin"] = data["x"] - w * (1 - hjust)
        data["xmax"] = data["x"] + w * hjust
        data["ymin"] = data["y"] - h * (1 - vjust)
        data["ymax"] = data["y"] + h * vjust
        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        from matplotlib.colors import to_rgba_array
        from matplotlib.image import AxesImage

        if not isinstance(coord, coord_cartesian):
            raise PlotnineError(
                "geom_raster only works with cartesian coordinates"
            )

        data = coord.transform(data, panel_params)
        x = data["x"].to_numpy().astype(float)
        y = data["y"].to_numpy().astype(float)
        facecolor = to_rgba_array(data["fill"].to_numpy())
        facecolor[:, 3] = data["alpha"].to_numpy()

        # Convert vector of data to flat image,
        # figure out dimensions of raster on plot, and the colored
        # indices.
        x_pos = ((x - x.min()) / resolution(x, False)).astype(int)
        y_pos = ((y - y.min()) / resolution(y, False)).astype(int)
        nrow = y_pos.max() + 1
        ncol = x_pos.max() + 1
        yidx, xidx = nrow - y_pos - 1, x_pos

        # Create and "color" the matrix.
        # Any gaps left whites (ones) colors pluse zero alpha values
        # allows makes it possible to have a "neutral" interpolation
        # into the gaps when intervals are uneven.
        X = np.ones((nrow, ncol, 4))
        X[:, :, 3] = 0
        X[yidx, xidx] = facecolor

        im = AxesImage(
            ax,
            data=X,
            interpolation=params["interpolation"],
            origin="upper",
            extent=(
                data["xmin"].min(),
                data["xmax"].max(),
                data["ymin"].min(),
                data["ymax"].max(),
            ),
            rasterized=params["raster"],
            filterrad=params["filterrad"],
            zorder=params["zorder"],
        )
        ax.add_image(im)



"""
Coordinates
"""
from .coord_cartesian import coord_cartesian
from .coord_fixed import coord_equal, coord_fixed
from .coord_flip import coord_flip
from .coord_trans import coord_trans

__all__ = [
    "coord_cartesian",
    "coord_fixed",
    "coord_equal",
    "coord_flip",
    "coord_trans",
]



"""
Little functions used all over the codebase
"""
from __future__ import annotations

import collections
import inspect
import itertools
import typing
import warnings
from contextlib import suppress
from warnings import warn
from weakref import WeakValueDictionary

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

# missing in type stubs
from pandas.core.groupby import DataFrameGroupBy  # type: ignore

from .exceptions import PlotnineError, PlotnineWarning
from .mapping import aes

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Sequence

    import numpy.typing as npt
    from IPython.core.interactiveshell import InteractiveShell
    from typing_extensions import TypeGuard

    from plotnine.typing import DataLike, FloatArray, FloatArrayLike, IntArray


# Points and lines of equal size should give the
# same visual diameter (for points) and thickness
# (for lines). Given the adjustments in geom_point,
# this factor gives us the match.
SIZE_FACTOR = np.sqrt(np.pi)


def is_scalar_or_string(val):
    """
    Return whether the given object is a scalar or string like.
    """
    return is_string(val) or not np.iterable(val)


def is_string(obj: Any) -> TypeGuard[str]:
    """
    Return True if *obj* is a string
    """
    return isinstance(obj, str)


def is_list_like(obj: Any) -> bool:
    """
    Return True if *obj* is a list, tuple, series or array
    """
    return isinstance(obj, (list, tuple, pd.Series, np.ndarray))


def identity(*args: Any) -> Any:
    """
    Return whatever is passed in
    """
    return args if len(args) > 1 else args[0]


def match(
    v1, v2, nomatch=-1, incomparables=None, start=0
) -> npt.NDArray[np.int64]:
    """
    Return a vector of the positions of (first)
    matches of its first argument in its second.

    Parameters
    ----------
    v1: array-like
        the values to be matched

    v2: array-like
        the values to be matched against

    nomatch: int
        the value to be returned in the case when
        no match is found.

    incomparables: array-like
        a list of values that cannot be matched.
        Any value in v1 matching a value in this list
        is assigned the nomatch value.
    start: int
        type of indexing to use. Most likely 0 or 1
    """
    # NOTE: This function gets called a lot. If it can
    # be optimised, it should.
    lookup = {}
    for i, x in enumerate(v2):
        if x not in lookup:
            lookup[x] = i

    if incomparables:
        skip = set(incomparables) if incomparables else set()
        lst = [
            lookup[x] + start if x not in skip and x in lookup else nomatch
            for x in v1
        ]
    else:
        lst = [lookup[x] + start if x in lookup else nomatch for x in v1]
    return np.array(lst)


def _margins(
    vars: tuple[list[str], list[str]], margins: bool | list[str] = True
):
    """
    Figure out margining variables.

    Given the variables that form the rows and
    columns, and a set of desired margins, works
    out which ones are possible. Variables that
    can't be margined over are dropped silently.

    Parameters
    ----------
    vars : list
        variable names for rows and columns
    margins : bool | list
        If true, margins over all vars, otherwise
        only those listed

    Return
    ------
    out : list
        All the margins to create.
    """
    if margins is False:
        return []

    def fn(_vars):
        "The margin variables for a given row or column"
        # The first item is and empty list for no margins
        dim_margins = [[]]
        # for each wanted variable, couple it with
        # all variables to the right
        for i, u in enumerate(_vars):
            if margins is True or u in margins:
                lst = [u] + list(_vars[i + 1 :])
                dim_margins.append(lst)
        return dim_margins

    # Margin variables for rows and columns
    row_margins = fn(vars[0])
    col_margins = fn(vars[1])

    # Cross the two
    lst = list(itertools.product(col_margins, row_margins))

    # Clean up -- merge the row and column variables
    pretty = []
    for c, r in lst:
        pretty.append(r + c)
    return pretty


def add_margins(
    df: pd.DataFrame,
    vars: tuple[list[str], list[str]],
    margins: bool | list[str] = True,
) -> pd.DataFrame:
    """
    Add margins to a data frame.

    All margining variables will be converted to factors.

    Parameters
    ----------
    df : dataframe
        input data frame

    vars : list
        a list of 2 lists | tuples vectors giving the
        variables in each dimension

    margins : bool | list
        variable names to compute margins for.
        True will compute all possible margins.
    """
    margin_vars = _margins(vars, margins)
    if not margin_vars:
        return df

    # create margin dataframes
    margin_dfs = [df]
    for vlst in margin_vars[1:]:
        dfx = df.copy()
        for v in vlst:
            dfx.loc[0:, v] = "(all)"
        margin_dfs.append(dfx)

    merged = pd.concat(margin_dfs, axis=0)
    merged.reset_index(drop=True, inplace=True)

    # All margin columns become categoricals. The margin indicator
    # (all) needs to be added as the last level of the categories.
    categories = {}
    for v in itertools.chain(*vars):
        col = df[v]
        if not pdtypes.is_categorical_dtype(df[v]):
            col = pd.Categorical(df[v])
        categories[v] = col.categories
        if "(all)" not in categories[v]:
            categories[v] = categories[v].insert(len(categories[v]), "(all)")

    for v in merged.columns.intersection(list(categories.keys())):
        merged[v] = merged[v].astype(pdtypes.CategoricalDtype(categories[v]))

    return merged


def ninteraction(df: pd.DataFrame, drop: bool = False) -> list[int]:
    """
    Compute a unique numeric id for each unique row in
    a data frame. The ids start at 1 -- in the spirit
    of `plyr::id`

    Parameters
    ----------
    df : dataframe
        Rows
    drop : bool
        If true, drop unused categorical levels leaving no
        gaps in the assignments.

    Returns
    -------
    out : list
        Row asssignments.

    Notes
    -----
    So far there has been no need not to drop unused levels
    of categorical variables.
    """
    if len(df) == 0:
        return []

    # Special case for single variable
    if len(df.columns) == 1:
        return _id_var(df[df.columns[0]], drop)

    # Calculate individual ids
    ids = df.apply(_id_var, axis=0)
    ids = ids.reindex(columns=list(reversed(ids.columns)))

    # Calculate dimensions
    def len_unique(x):
        return len(np.unique(x))

    ndistinct: IntArray = ids.apply(len_unique, axis=0).to_numpy()

    combs = np.array(np.hstack([1, np.cumprod(ndistinct[:-1])]))
    mat = np.array(ids)
    res = (mat - 1) @ combs.T + 1
    res = np.array(res).flatten().tolist()

    if drop:
        return _id_var(res, drop)
    else:
        return list(res)


def _id_var(x: pd.Series[Any], drop: bool = False) -> list[int]:
    """
    Assign ids to items in x. If two items
    are the same, they get the same id.

    Parameters
    ----------
    x : array-like
        items to associate ids with
    drop : bool
        Whether to drop unused factor levels
    """
    if len(x) == 0:
        return []

    categorical = pdtypes.is_categorical_dtype(x)

    if categorical:
        if drop:
            x = x.cat.remove_unused_categories()
            lst = list(x.cat.codes + 1)
        else:
            has_nan = any(np.isnan(i) for i in x if isinstance(i, float))
            if has_nan:
                # NaNs are -1, we give them the highest code
                nan_code = -1
                new_nan_code = np.max(x.cat.codes) + 1
                lst = [val if val != nan_code else new_nan_code for val in x]
            else:
                lst = list(x.cat.codes + 1)
    else:
        try:
            levels = sorted(set(x))
        except TypeError:
            from mizani.utils import multitype_sort

            # x probably has NANs
            levels = multitype_sort(set(x))

        lst = match(x, levels)
        lst = [item + 1 for item in lst]

    return lst


def join_keys(x, y, by=None):
    """
    Join keys.

    Given two data frames, create a unique key for each row.

    Parameters
    -----------
    x : dataframe
    y : dataframe
    by : list-like
        Column names to join by

    Returns
    -------
    out : dict
        Dictionary with keys x and y. The values of both keys
        are arrays with integer elements. Identical rows in
        x and y dataframes would have the same key in the
        output. The key elements start at 1.
    """
    if by is None:
        by = slice(None, None, None)

    if isinstance(by, tuple):
        by = list(by)

    joint = pd.concat([x[by], y[by]], ignore_index=True)
    keys = ninteraction(joint, drop=True)
    keys = np.asarray(keys)
    nx, ny = len(x), len(y)
    return {"x": keys[np.arange(nx)], "y": keys[nx + np.arange(ny)]}


def check_required_aesthetics(required, present, name):
    missing_aes = set(required) - set(present)

    if missing_aes:
        msg = "{} requires the following missing aesthetics: {}"
        raise PlotnineError(msg.format(name, ", ".join(missing_aes)))


def uniquecols(df):
    """
    Return unique columns

    This is used for figuring out which columns are
    constant within a group
    """
    bool_idx = df.apply(lambda col: len(np.unique(col)) == 1, axis=0)
    df = df.loc[:, bool_idx].iloc[0:1, :].reset_index(drop=True)
    return df


def jitter(x, factor=1, amount=None, random_state=None):
    """
    Add a small amount of noise to values in an array_like

    Parameters
    ----------
    x : array_like
        Values to apply a jitter
    factor : float
        Multiplicative value to used in automatically determining
        the `amount`. If the `amount` is given then the `factor`
        has no effect.
    amount : float
        This defines the range ([-amount, amount]) of the jitter to
        apply to the values. If `0` then ``amount = factor * z/50``.
        If `None` then ``amount = factor * d/5``, where d is about
        the smallest difference between `x` values and `z` is the
        range of the `x` values.
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.

    References:

        - Chambers, J. M., Cleveland, W. S., Kleiner, B. and Tukey,
          P.A. (1983) *Graphical Methods for Data Analysis*. Wadsworth;
          figures 2.8, 4.22, 5.4.
    """
    if len(x) == 0:
        return x

    if random_state is None:
        random_state = np.random
    elif isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)

    x = np.asarray(x)

    try:
        z = np.ptp(x[np.isfinite(x)])
    except IndexError:
        z = 0

    if z == 0:
        z = np.abs(np.min(x))
    if z == 0:
        z = 1

    if amount is None:
        _x = np.round(x, 3 - int(np.floor(np.log10(z)))).astype(int)
        xx = np.unique(np.sort(_x))
        d = np.diff(xx)
        if len(d):
            d = d.min()
        elif xx != 0:
            d = xx / 10.0
        else:
            d = z / 10
        amount = factor / 5.0 * abs(d)
    elif amount == 0:
        amount = factor * (z / 50.0)

    return x + random_state.uniform(-amount, amount, len(x))


def remove_missing(
    df: pd.DataFrame,
    na_rm: bool = False,
    vars: Sequence[str] | None = None,
    name: str = "",
    finite: bool = False,
) -> pd.DataFrame:
    """
    Convenience function to remove missing values from a dataframe

    Parameters
    ----------
    df : dataframe
    na_rm : bool
        If False remove all non-complete rows with and show warning.
    vars : list-like
        columns to act on
    name : str
        Name of calling method for a more informative message
    finite : bool
        If True replace the infinite values in addition to the NaNs
    """
    n = len(df)

    if vars is None:
        vars = df.columns.to_list()
    else:
        vars = df.columns.intersection(list(vars)).to_list()

    if finite:
        lst = [np.inf, -np.inf]
        to_replace = {v: lst for v in vars}
        df.replace(to_replace, np.nan, inplace=True)
        txt = "non-finite"
    else:
        txt = "missing"

    df = df.dropna(subset=vars)
    df.reset_index(drop=True, inplace=True)
    if len(df) < n and not na_rm:
        msg = "{} : Removed {} rows containing {} values."
        warn(msg.format(name, n - len(df), txt), PlotnineWarning, stacklevel=3)
    return df


def to_rgba(colors, alpha):
    """
    Covert hex colors to rgba values.

    Parameters
    ----------
    colors : iterable | str
        colors to convert
    alphas : iterable | float
        alpha values

    Returns
    -------
    out : ndarray | tuple
        rgba color(s)

    Notes
    -----
    Matplotlib plotting functions only accept scalar
    alpha values. Hence no two objects with different
    alpha values may be plotted in one call. This would
    make plots with continuous alpha values innefficient.
    However :), the colors can be rgba hex values or
    list-likes and the alpha dimension will be respected.
    """

    def is_iterable(var):
        return np.iterable(var) and not is_string(var)

    def has_alpha(c):
        if isinstance(c, tuple):
            if len(c) == 4:
                return True
        elif isinstance(c, str):
            if c[0] == "#" and len(c) == 9:
                return True
        return False

    def no_color(c):
        return c is None or c == "" or c.lower() == "none"

    def to_rgba_hex(c, a):
        """
        Conver rgb color to rgba hex value

        If color c has an alpha channel, then alpha value
        a is ignored
        """
        from matplotlib.colors import colorConverter, to_hex

        if c in ("None", "none"):
            return c

        _has_alpha = has_alpha(c)
        c = to_hex(c, keep_alpha=_has_alpha)

        if not _has_alpha:
            arr = colorConverter.to_rgba(c, a)
            return to_hex(arr, keep_alpha=True)

        return c

    if is_iterable(colors):
        if all(no_color(c) for c in colors):
            return "none"

        if is_iterable(alpha):
            return [to_rgba_hex(c, a) for c, a in zip(colors, alpha)]
        else:
            return [to_rgba_hex(c, alpha) for c in colors]
    else:
        if no_color(colors):
            return colors
        if is_iterable(alpha):
            return [to_rgba_hex(colors, a) for a in alpha]
        else:
            return to_rgba_hex(colors, alpha)


def groupby_apply(
    df: pd.DataFrame,
    cols: str | list[str],
    func: Callable[..., pd.DataFrame],
    *args: tuple[Any],
    **kwargs: Any,
) -> pd.DataFrame:
    """
    Groupby cols and call the function fn on each grouped dataframe.

    Parameters
    ----------
    cols : str | list of str
        columns to groupby
    func : function
        function to call on the grouped data
    *args : tuple
        positional parameters to pass to func
    **kwargs : dict
        keyword parameter to pass to func

    This is meant to avoid pandas df.groupby('col').apply(fn, *args),
    as it calls fn twice on the first dataframe. If the nested code also
    does the same thing, it can be very expensive
    """
    if df.empty:
        return df.copy()

    try:
        axis = kwargs.pop("axis")
    except KeyError:
        axis = 0

    lst = []
    for _, d in df.groupby(cols):
        # function fn should be free to modify dataframe d, therefore
        # do not mark d as a slice of df i.e no SettingWithCopyWarning
        lst.append(func(d, *args, **kwargs))
    return pd.concat(lst, axis=axis, ignore_index=True)


def pivot_apply(df, column, index, func, *args, **kwargs):
    """
    Apply a function to each group of a column

    The function is kind of equivalent to R's *tapply*.

    Parameters
    ----------
    df : dataframe
        Dataframe to be pivoted
    column : str
        Column to apply function to.
    index : str
        Column that will be grouped on (and whose unique values
        will make up the index of the returned dataframe)
    func : function
        Function to apply to each column group. It *should* return
        a single value.
    *args : tuple
        Arguments to ``func``
    **kwargs : dict
        Keyword arguments to ``func``

    Returns
    -------
    out : dataframe
        Dataframe with index ``index`` and column ``column`` of
        computed/aggregate values .
    """

    def _func(x):
        return func(x, *args, **kwargs)

    return df.pivot_table(column, index, aggfunc=_func)[column]


def make_line_segments(
    x: FloatArrayLike, y: FloatArrayLike, ispath=True
) -> FloatArray:
    """
    Return an (n x 2 x 2) array of n line segments

    Parameters
    ----------
    x : array-like
        x points
    y : array-like
        y points
    ispath : bool
        If True, the points represent a path from one point
        to the next until the last. If False, then each pair
        of successive(even-odd pair) points yields a line.
    """
    # Series objects would otherwise require .iloc
    x = np.asarray(x)
    y = np.asarray(y)
    if ispath:
        x = interleave(x[:-1], x[1:])
        y = interleave(y[:-1], y[1:])
    elif len(x) % 2:
        raise PlotnineError("Expects an even number of points")

    n = len(x) // 2
    segments = np.reshape(list(zip(x, y)), [n, 2, 2])
    return segments


def copy_keys(source, destination, keys=None):
    """
    Add keys in source to destination

    Parameters
    ----------
    source : dict

    destination: dict

    keys : None | iterable
        The keys in source to be copied into destination. If
        None, then `keys = destination.keys()`
    """
    if keys is None:
        keys = destination.keys()
    for k in set(source) & set(keys):
        destination[k] = source[k]
    return destination


class RegistryMeta(type):
    """
    Make a metaclass scriptable
    """

    _registry: WeakValueDictionary[str, Any] = WeakValueDictionary()

    # In here "self" refers to the meta class although it is never
    # instantiated
    def __getitem__(self, key):
        try:
            return self._registry[key]
        except KeyError:
            msg = (
                "'{}' Not in Registry. Make sure the module in "
                "which it is defined has been imported."
            )
            raise PlotnineError(msg.format(key))

    def __setitem__(self, key, value):
        self._registry[key] = value

    def __iter__(self):
        return self._registry.__iter__()

    def keys(self):
        return self._registry.keys()

    def values(self):
        return self._registry.values()

    def items(self):
        return self._registry.items()


class Registry(type, metaclass=RegistryMeta):
    """
    Creates class that automatically registers all subclasses

    To prevent the base class from showing up in the registry,
    it should have an attribute `__base__ = True`. This metaclass
    uses a single dictionary to register all types of subclasses.

    To access the registered objects, use:

        obj = Registry['name']

    To explicitly register objects

        Registry['name'] = obj

    Notes
    -----
    When objects are deleted, they are automatically removed
    from the Registry.
    """

    # namespace is of the class that subclasses Registry (or a
    # subclasses a subclass of Registry, ...) being created
    # e.g. geom, geom_point, ...
    def __new__(cls, name, bases, namespace):
        sub_cls = super().__new__(cls, name, bases, namespace)
        if not namespace.pop("__base__", False):
            cls._registry[name] = sub_cls
        return sub_cls


class RegistryHierarchyMeta(type):
    """
    Create a class that registers subclasses and the Hierarchy

    The class has gets two properties:

    1. ``_registry`` a dictionary of all the subclasses of the
       base class. The keys are the names of the classes and
       the values are the class objects.
    2. ``_hierarchy`` a dictionary (default) that holds the
       inheritance hierarchy of each class. Each key is a class
       and the value is a list of classes. The first name in the
       list is that of the key class.

    The goal of the ``_hierarchy`` object to facilitate the
    lookup of themeable properties taking into consideration the
    inheritance hierarchy. For example if `strip_text_x` inherits
    from `strip_text` which inherits from `text`, then if a property
    of `strip_text_x` is requested, the lookup should fallback to
    the other two if `strip_text_x` is not present or is missing
    the requested property.
    """

    def __init__(cls, name, bases, namespace):
        if not hasattr(cls, "_registry"):
            cls._registry = {}
            cls._hierarchy = collections.defaultdict(list)
        else:
            cls._registry[name] = cls
            cls._hierarchy[name].append(name)
            for base in bases:
                for base2 in base.mro()[:-2]:
                    cls._hierarchy[base2.__name__].append(name)


def alias(name, class_object):
    """
    Create an alias of a class object

    The objective of this method is to have
    an alias that is Registered. i.e If we have

        class_b = class_a

    Makes `class_b` an alias of `class_a`, but if
    `class_a` is registered by its metaclass,
    `class_b` is not. The solution

        alias('class_b', class_a)

    is equivalent to:

        class_b = class_a
        Register['class_b'] = class_a
    """
    module = inspect.getmodule(class_object)
    module.__dict__[name] = class_object
    if isinstance(class_object, Registry):
        Registry[name] = class_object


def get_kwarg_names(func):
    """
    Return a list of valid kwargs to function func
    """
    sig = inspect.signature(func)
    kwonlyargs = [
        p.name for p in sig.parameters.values() if p.default is not p.empty
    ]
    return kwonlyargs


def get_valid_kwargs(func, potential_kwargs):
    """
    Return valid kwargs to function func
    """
    kwargs = {}
    for name in get_kwarg_names(func):
        with suppress(KeyError):
            kwargs[name] = potential_kwargs[name]
    return kwargs


def copy_missing_columns(df, ref_df):
    """
    Copy missing columns from ref_df to df

    If df and ref_df are the same length, the columns are
    copied in the entirety. If the length ofref_df is a
    divisor of the length of df, then the values of the
    columns from ref_df are repeated.

    Otherwise if not the same length, df gets a column
    where all elements are the same as the first element
    in ref_df

    Parameters
    ----------
    df : dataframe
        Dataframe to which columns will be added
    ref_df : dataframe
        Dataframe from which columns will be copied
    """
    cols = ref_df.columns.difference(df.columns)
    _loc = ref_df.columns.get_loc

    l1, l2 = len(df), len(ref_df)
    if l1 >= l2 and l1 % l2 == 0:
        idx = np.tile(range(l2), l1 // l2)
    else:
        idx = np.repeat(0, l1)

    for col in cols:
        df[col] = ref_df.iloc[idx, _loc(col)].values


def data_mapping_as_kwargs(args, kwargs):
    """
    Return kwargs with the mapping and data values

    Parameters
    ----------
    args : tuple
        Arguments to :class:`geom` or :class:`stat`.
    kwargs : dict
        Keyword arguments to :class:`geom` or :class:`stat`.

    Returns
    -------
    out : dict
        kwargs that includes 'data' and 'mapping' keys.
    """
    data, mapping = order_as_data_mapping(*args)

    # check kwargs #
    if mapping is not None:
        if "mapping" in kwargs:
            raise PlotnineError("More than one mapping argument.")
        else:
            kwargs["mapping"] = mapping
    else:
        if "mapping" not in kwargs:
            mapping = aes()

    if kwargs.get("mapping", None) is None:
        kwargs["mapping"] = mapping

    if data is not None and "data" in kwargs:
        raise PlotnineError("More than one data argument.")
    elif "data" not in kwargs:
        kwargs["data"] = data

    duplicates = set(kwargs["mapping"]) & set(kwargs)
    if duplicates:
        msg = "Aesthetics {} specified two times."
        raise PlotnineError(msg.format(duplicates))
    return kwargs


def ungroup(data: DataLike) -> DataLike:
    """Return an ungrouped DataFrame, or pass the original data back."""

    if isinstance(data, DataFrameGroupBy):
        return data.obj  # type: ignore

    return data


def order_as_data_mapping(
    arg1: DataLike | aes | None,
    arg2: DataLike | aes | None,
) -> tuple[DataLike | None, aes | None]:
    """
    Reorder args to ensure (data, mapping) order

    This function allow the user to pass mapping and data
    to ggplot and geom in any order.

    Parameter
    ---------
    arg1 : pd.DataFrame | aes
        Dataframe or aes Mapping
    arg2 : pd.DataFrame | aes
        Dataframe or aes Mapping

    Returns
    -------
    data : pd.DataFrame | callable
    mapping : aes
    """
    # Valid types for the values are:
    #   - None, None
    #   - Dataframe, None
    #   - None, DataFrame
    #   - aes, None
    #   - None, aes
    #   - DataFrame, aes
    data: DataLike | None = None
    mapping: aes | None = None

    for arg in [arg1, arg2]:
        if isinstance(arg, aes):
            if mapping is None:
                mapping = arg
            else:
                raise TypeError(
                    "Expected a single aesthetic mapping, found two"
                )
        elif is_data_like(arg):
            if data is None:
                data = arg
            else:
                raise TypeError("Expected a single dataframe, found two")
        elif arg is not None:
            raise TypeError(
                f"Bad type of argument {arg!r}, expected a dataframe "
                "or a mapping."
            )

    return data, mapping


# Returning a type guard here is not fully sound, because if `obj`
# is a callable, we aren't checking that it has no required args
# and we can't check the return value's type.
def is_data_like(obj: Any) -> TypeGuard[DataLike]:
    """
    Return True if obj could be data

    Parameters
    ----------
    obj : object
        Object that could be data

    Returns
    -------
    out : bool
        Whether obj could represent data as expected by
        ggplot(), geom() or stat().
    """
    return (
        isinstance(obj, pd.DataFrame)
        or callable(obj)
        or hasattr(obj, "to_pandas")
    )


def interleave(*arrays):
    """
    Interleave arrays

    All arrays/lists must be the same length

    Parameters
    ----------
    arrays : tup
        2 or more arrays to interleave

    Return
    ------
    out : np.array
        Result from interleaving the input arrays
    """
    return np.column_stack(arrays).ravel()


def resolution(x, zero=True):
    """
    Compute the resolution of a data vector

    Resolution is smallest non-zero distance between adjacent values

    Parameters
    ----------
    x    : 1D array-like
    zero : Boolean
        Whether to include zero values in the computation

    Result
    ------
    res : resolution of x
        If x is an integer array, then the resolution is 1
    """
    from mizani.bounds import zero_range

    x = np.asarray(x)

    # (unsigned) integers or an effective range of zero
    _x = x[~pd.isna(x)]
    _x = (x.min(), x.max())
    if x.dtype.kind in ("i", "u") or zero_range(_x):
        return 1

    x = np.unique(x)
    if zero:
        x = np.unique(np.hstack([0, x]))

    return np.min(np.diff(np.sort(x)))


def cross_join(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    """
    Return a dataframe that is a cross between dataframes
    df1 and df2

    ref: https://github.com/pydata/pandas/issues/5401
    """
    if len(df1) == 0:
        return df2

    if len(df2) == 0:
        return df1

    # Add as lists so that the new index keeps the items in
    # the order that they are added together
    all_columns = list(pd.Index(list(df1.columns) + list(df2.columns)))
    df1["key"] = 1
    df2["key"] = 1
    return pd.merge(df1, df2, on="key").loc[:, all_columns]


def to_inches(value: float, units: str) -> float:
    """
    Convert value to inches

    Parameters
    ----------
    value : float
        Value to be converted
    units : str
        Units of value. Must be one of
        `['in', 'cm', 'mm']`.
    """
    lookup: dict[str, Callable[[float], float]] = {
        "in": lambda x: x,
        "cm": lambda x: x / 2.54,
        "mm": lambda x: x / (2.54 * 10),
    }
    try:
        return lookup[units](value)
    except KeyError:
        raise PlotnineError(f"Unknown units '{units}'")


def from_inches(value: float, units: str) -> float:
    """
    Convert value in inches to given units

    Parameters
    ----------
    value : float
        Value to be converted
    units : str
        Units to convert value to. Must be one of
        `['in', 'cm', 'mm']`.
    """
    lookup: dict[str, Callable[[float], float]] = {
        "in": lambda x: x,
        "cm": lambda x: x * 2.54,
        "mm": lambda x: x * 2.54 * 10,
    }
    try:
        return lookup[units](value)
    except KeyError:
        raise PlotnineError(f"Unknown units '{units}'")


class array_kind:
    @staticmethod
    def discrete(arr):
        """
        Return True if array is discrete

        Parameters
        ----------
        arr : numpy.array
            Must have a dtype

        Returns
        -------
        out : bool
            Whether array `arr` is discrete
        """
        return arr.dtype.kind in "ObUS"

    @staticmethod
    def continuous(arr):
        """
        Return True if array is continuous

        Parameters
        ----------
        arr : numpy.array or pandas.series
            Must have a dtype

        Returns
        -------
        out : bool
            Whether array `arr` is continuous
        """
        return arr.dtype.kind in "ifuc"

    @staticmethod
    def datetime(arr):
        return arr.dtype.kind == "M"

    @staticmethod
    def timedelta(arr):
        return arr.dtype.kind == "m"

    @staticmethod
    def ordinal(arr):
        """
        Return True if array is an ordered categorical

        Parameters
        ----------
        arr : numpy.array
            Must have a dtype

        Returns
        -------
        out : bool
            Whether array `arr` is an ordered categorical
        """
        if pdtypes.is_categorical_dtype(arr):
            return arr.cat.ordered
        return False


def log(x, base=None):
    """
    Calculate the log

    Parameters
    ----------
    x : float or array_like
        Input values
    base : int or float (Default: None)
        Base of the log. If `None`, the natural logarithm
        is computed (`base=np.e`).

    Returns
    -------
    out : float or ndarray
        Calculated result
    """
    if base == 10:
        return np.log10(x)
    elif base == 2:
        return np.log2(x)
    elif base is None or base == np.e:
        return np.log(x)
    else:
        return np.log(x) / np.log(base)


class ignore_warnings:
    """
    Ignore Warnings Context Manager

    Wrap around warnings.catch_warnings to make ignoring
    warnings easier.

    Parameters
    ----------
    *categories : tuple
        Warning categories to ignore e.g UserWarning,
        FutureWarning, RuntimeWarning, ...
    """

    _cm: warnings.catch_warnings

    def __init__(self, *categories):
        self.categories = categories
        self._cm = warnings.catch_warnings()

    def __enter__(self):
        self._cm.__enter__()
        for c in self.categories:
            warnings.filterwarnings("ignore", category=c)

    def __exit__(self, type, value, traceback):
        return self._cm.__exit__(type, value, traceback)


def get_ipython() -> "InteractiveShell" | None:
    """
    Return running IPython instance or None
    """
    try:
        from IPython.core.getipython import get_ipython
    except ImportError:
        return None
    return get_ipython()



from copy import copy, deepcopy

from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme
from .theme_gray import theme_gray


class theme_xkcd(theme_gray):
    """
    xkcd theme

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 12.
    scale : float, optional
        The amplitude of the wiggle perpendicular to the line
        (in pixels). Default is 1.
    length : float, optional
        The length of the wiggle along the line (in pixels).
        Default is 100.
    randomness : float, optional
        The factor by which the length is randomly scaled.
        Default is 2.
    stroke_size : float, optional
        Size of the stroke to apply to the lines and text paths.
        Default is 4.
    stroke_color : str or tuple, optional
        Color of the strokes. Default is ``white``.
        For no color, use ``'none'``.
    """

    def __init__(
        self,
        base_size=12,
        scale=1,
        length=100,
        randomness=2,
        stroke_size=4,
        stroke_color="white",
    ):
        from matplotlib import patheffects

        super().__init__(base_size)
        self += theme(
            text=element_text(family=["xkcd", "Humor Sans", "Comic Sans MS"]),
            axis_ticks=element_line(color="black", size=1.5),
            axis_ticks_minor=element_blank(),
            axis_ticks_direction="in",
            axis_ticks_length_major=6,
            legend_background=element_rect(color="black", fill="None"),
            legend_margin=5,
            legend_key=element_rect(fill="None"),
            panel_border=element_rect(color="black", size=1.5),
            panel_grid=element_blank(),
            panel_background=element_rect(fill="white"),
            strip_background=element_rect(color="black", fill="white"),
            strip_background_x=element_rect(width=2 / 3.0),
            strip_background_y=element_rect(height=2 / 3.0),
            strip_align=-0.5,
        )

        d = {
            "axes.unicode_minus": False,
            "path.sketch": (scale, length, randomness),
            "path.effects": [
                patheffects.withStroke(
                    linewidth=stroke_size, foreground=stroke_color
                )
            ],
        }
        self._rcParams.update(d)

    def __deepcopy__(self, memo):
        """
        Deep copy support for theme_xkcd
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        for key, item in old.items():
            if key == "_rcParams":
                continue
            new[key] = deepcopy(old[key], memo)

        result._rcParams = {}
        for k, v in self._rcParams.items():
            try:
                result._rcParams[k] = deepcopy(v, memo)
            except NotImplementedError:
                # deepcopy raises an error for objects that are
                # drived from or composed of
                # matplotlib.transform. TransformNode.
                # Not desirable, but probably requires upstream fix.
                # In particular, XKCD uses
                # matplotlib.patheffects.withStrok
                # -gdowding
                result._rcParams[k] = copy(v)

        return result



import string

import numpy as np
import pandas as pd

from plotnine import (
    aes,
    coord_equal,
    geom_point,
    ggplot,
    guides,
    lims,
    scale_color_identity,
    scale_shape_manual,
)


def test_aesthetics():
    df = pd.DataFrame(
        {
            "a": range(5),
            "b": 2,
            "c": 3,
            "d": 4,
            "e": 5,
            "f": 6,
            "g": 7,
            "h": 8,
            "i": 9,
        }
    )

    p = (
        ggplot(df, aes(y="a"))
        + geom_point(aes(x="b"))
        + geom_point(aes(x="c", size="a"))
        + geom_point(aes(x="d", alpha="a"), size=10, show_legend=False)
        + geom_point(aes(x="e", shape="factor(a)"), size=10, show_legend=False)
        + geom_point(aes(x="f", color="factor(a)"), size=10, show_legend=False)
        + geom_point(
            aes(x="g", fill="a"), stroke=0, size=10, show_legend=False
        )
        + geom_point(
            aes(x="h", stroke="a"), fill="white", color="green", size=10
        )
        + geom_point(
            aes(x="i", shape="factor(a)"),
            fill="brown",
            stroke=2,
            size=10,
            show_legend=False,
        )
    )

    assert p == "aesthetics"


def test_no_fill():
    df = pd.DataFrame({"x": range(5), "y": range(5)})

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point(color="red", fill=None, size=5, stroke=1.5)
        + geom_point(
            aes(y="y+1"), color="blue", fill="none", size=5, stroke=1.5
        )
        + geom_point(aes(y="y+2"), color="green", fill="", size=5, stroke=1.5)
        + geom_point(
            aes(y="y+3"), color="yellow", fill="gray", size=5, stroke=1.5
        )
    )

    assert p == "no_fill"


def test_legend_transparency():
    n = 5

    data = pd.DataFrame(
        {
            "x": list(range(n)) * 3,
            "y": [1] * n + [2] * n + [3] * n,
            "color": ["orange"] * n + ["red"] * n + ["#0000FF44"] * n,
        }
    )

    p = (
        ggplot(data, aes("x", "y", color="color"))
        + geom_point(size=25, stroke=3)
        + lims(x=(-2.5, 6.5), y=(-1.5, 5.5))
        + scale_color_identity(guide="legend")
    )
    assert p == "legend_transparency"


class TestColorFillonUnfilledShape:
    df = pd.DataFrame({"x": range(6), "y": range(6), "z": list("aabbcc")})
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point(shape="3", size=10, stroke=3)
        + guides(fill=False)
    )

    # Color  Fill  Result
    # No     No    Black
    # No     Yes   Black
    # Yes    No    Color
    # Yes    Yes   Color

    def test_no_mapping(self):
        assert self.p == "no_mapping"

    def test_fill_only_mapping(self):
        p = self.p + aes(fill="x")
        # Same as above
        assert p == "no_mapping"

    def test_color_only_mapping(self):
        p = self.p + aes(color="z")
        assert p == "color_only_mapping"

    def test_color_fill_mapping(self):
        p = self.p + aes(color="z", fill="x")
        # Same as above
        assert p == "color_only_mapping"


def test_custom_shapes():
    n = 26
    shapes = [rf"$\mathrm{{{x}}}$" for x in string.ascii_uppercase]
    theta = np.linspace(0, 2 * np.pi * (1 - 1 / n), n)

    df = pd.DataFrame(
        {
            "x": np.sin(theta),
            "y": np.cos(theta),
            "theta": theta,
        }
    )

    p = (
        ggplot(df, aes("x", "y", shape="factor(range(n))", color="theta"))
        + geom_point(size=10, show_legend=False)
        + scale_shape_manual(shapes)
        + coord_equal()
    )
    assert p == "custom_shapes"



from __future__ import annotations

import typing
from typing import overload

import pandas as pd

from ..iapi import labels_view, panel_ranges, panel_view
from .coord_cartesian import coord_cartesian

if typing.TYPE_CHECKING:
    from plotnine.typing import Scale


class coord_flip(coord_cartesian):
    """
    Flipped cartesian coordinates

    The horizontal becomes vertical, and vertical becomes
    horizontal. This is primarily useful for converting
    geoms and statistics which display y conditional
    on x, to x conditional on y.

    Parameters
    ----------
    xlim : None | (float, float)
        Limits for x axis. If None, then they are
        automatically computed.
    ylim : None | (float, float)
        Limits for y axis. If None, then they are
        automatically computed.
    expand : bool
        If `True`, expand the coordinate axes by
        some factor. If `False`, use the limits
        from the data.
    """

    def labels(self, cur_labels: labels_view) -> labels_view:
        return flip_labels(super().labels(cur_labels))

    def transform(
        self, data: pd.DataFrame, panel_params: panel_view, munch: bool = False
    ) -> pd.DataFrame:
        data = flip_labels(data)
        return super().transform(data, panel_params, munch=munch)

    def setup_panel_params(self, scale_x: Scale, scale_y: Scale) -> panel_view:
        panel_params = super().setup_panel_params(scale_x, scale_y)
        return flip_labels(panel_params)

    def setup_layout(self, layout: pd.DataFrame) -> pd.DataFrame:
        # switch the scales
        x, y = "SCALE_X", "SCALE_Y"
        layout[x], layout[y] = layout[y].copy(), layout[x].copy()
        return layout

    def range(self, panel_params: panel_view) -> panel_ranges:
        """
        Return the range along the dimensions of the coordinate system
        """
        # Defaults to providing the 2D x-y ranges
        return panel_ranges(x=panel_params.y.range, y=panel_params.x.range)


@overload
def flip_labels(obj: pd.DataFrame) -> pd.DataFrame:
    ...


@overload
def flip_labels(obj: labels_view) -> labels_view:
    ...


@overload
def flip_labels(obj: panel_view) -> panel_view:
    ...


def flip_labels(
    obj: pd.DataFrame | labels_view | panel_view,
) -> pd.DataFrame | labels_view | panel_view:
    """
    Rename fields x to y and y to x

    Parameters
    ----------
    obj : dict_like | dataclass
        Object with labels to rename
    """

    def sub(a: str, b: str, df: pd.DataFrame):
        """
        Substitute all keys that start with a to b
        """
        columns: list[str] = df.columns.tolist()
        for label in columns:
            if label.startswith(a):
                new_label = b + label[1:]
                df[new_label] = df.pop(label)

    # TODO: pyright cannot handle
    # isinstance(obj, (labels_view, panel_view))
    # keep checking with future versions.
    if isinstance(obj, pd.DataFrame):
        sub("x", "z", obj)
        sub("y", "x", obj)
        sub("z", "y", obj)
    elif isinstance(obj, labels_view):
        obj.x, obj.y = obj.y, obj.x
    elif isinstance(obj, panel_view):
        obj.x, obj.y = obj.y, obj.x

    return obj



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    geom_freqpoly,
    geom_histogram,
    geom_point,
    ggplot,
)

n = 10  # Some even number greater than 2

# ladder: 0 1 times, 1 2 times, 2 3 times, ...
df = pd.DataFrame(
    {
        "x": np.repeat(range(n + 1), range(n + 1)),
        "z": np.repeat(range(n // 2), range(3, n * 2, 4)),
    }
)


def test_midpoint():
    p = (
        ggplot(df, aes("x"))
        + geom_histogram(aes(fill="factor(z)"), bins=n, alpha=0.25)
        + geom_freqpoly(bins=n, size=4)
        + geom_point(stat="bin", bins=n, size=4, stroke=0, color="red")
    )

    assert p == "midpoint"



import sys
from contextlib import suppress

import pandas as pd

from ..exceptions import PlotnineError
from ..geoms import geom_blank
from ..mapping.aes import ALL_AESTHETICS, aes
from ..scales.scales import make_scale
from ..utils import array_kind


# By adding limits, we create a scale of the appropriate type
class _lim:
    aesthetic = None

    def __init__(self, *limits):
        if not limits:
            msg = "{}lim(), is missing limits"
            raise PlotnineError(msg.format(self.aesthetic))
        elif len(limits) == 1:
            limits = limits[0]

        series = pd.Series(limits)

        # Type of transform
        if not any(x is None for x in limits) and limits[0] > limits[1]:
            self.trans = "reverse"
        elif array_kind.continuous(series):
            self.trans = "identity"
        elif array_kind.discrete(series):
            self.trans = None
        elif array_kind.datetime(series):
            self.trans = "datetime"
        elif array_kind.timedelta(series):
            self.trans = "timedelta"
        else:
            msg = f"Unknown type {type(limits[0])} of limits"
            raise TypeError(msg)

        self.limits = limits
        self.limits_series = series

    def get_scale(self, gg):
        """
        Create a scale
        """
        # This method does some introspection to save users from
        # scale mismatch error. This could happen when the
        # aesthetic is mapped to a categorical but the limits
        # are not provided in categorical form. We only handle
        # the case where the mapping uses an expression to
        # conver to categorical e.g `aes(color='factor(cyl)')`.
        # However if `'cyl'` column is a categorical and the
        # mapping is `aes(color='cyl')`, that will result in
        # an error. If later case proves common enough then we
        # could inspect the data and be clever based on that too!!
        ae = self.aesthetic
        series = self.limits_series
        ae_values = []

        # Look through all the mappings for this aesthetic,
        # if we detect any factor stuff then we convert the
        # limits data to categorical so that the right scale
        # can be choosen. This should take care of the most
        # common use cases.
        for layer in gg.layers:
            with suppress(KeyError):
                value = layer.mapping[ae]
                if isinstance(value, str):
                    ae_values.append(value)

        for value in ae_values:
            if "factor(" in value or "Categorical(" in value:
                series = pd.Categorical(self.limits_series)
                break
        return make_scale(
            self.aesthetic, series, limits=self.limits, trans=self.trans
        )

    def __radd__(self, gg):
        scale = self.get_scale(gg)
        gg.scales.append(scale)
        return gg


class xlim(_lim):
    """
    Set x-axis limits

    Parameters
    ----------
    limits : array_like
        Min and max limits. Must be of size 2.
        You can also pass two values e.g
        ``xlim(40, 100)``
    """

    aesthetic = "x"


class ylim(_lim):
    """
    Set y-axis limits

    Parameters
    ----------
    limits : array_like
        Min and max limits. Must be of size 2.
        You can also pass two values e.g
        ``ylim(40, 100)``

    Notes
    -----
    If the 2nd value of ``limits`` is less than
    the first, a reversed scale will be created.
    """

    aesthetic = "y"


class alphalim(_lim):
    """
    Alpha limits
    """

    aesthetic = "alpha"


class colorlim(_lim):
    """
    Color limits
    """

    aesthetic = "color"


class filllim(_lim):
    """
    Fill limits
    """

    aesthetic = "fill"


class linetypelim(_lim):
    """
    Linetype limits
    """

    aesthetic = "linetype"


class shapelim(_lim):
    """
    Shapee limits
    """

    aesthetic = "shape"


class sizelim(_lim):
    """
    Size limits
    """

    aesthetic = "size"


class strokelim(_lim):
    """
    Stroke limits
    """

    aesthetic = "stroke"


class lims:
    """
    Set aesthetic limits

    Parameters
    ----------
    kwargs : dict
        Aesthetic and the values of the limits.
        e.g ``x=(40, 100)``

    Notes
    -----
    If the 2nd value of ``limits`` is less than
    the first, a reversed scale will be created.
    """

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def __radd__(self, gg):
        """
        Add limits to ggplot object
        """
        thismodule = sys.modules[__name__]
        for ae, value in self._kwargs.items():
            try:
                klass = getattr(thismodule, f"{ae}lim")
            except AttributeError:
                raise PlotnineError("Cannot change limits for '{}'")

            gg += klass(value)

        return gg


def expand_limits(**kwargs):
    """
    Expand the limits any aesthetic using data

    Parameters
    ----------
    kwargs : dict or dataframe
        Data to use in expanding the limits.
        The keys should be aesthetic names
        e.g. *x*, *y*, *colour*, ...
    """

    def as_list(key):
        with suppress(KeyError):
            if isinstance(kwargs[key], (int, float, str)):
                kwargs[key] = [kwargs[key]]

    if isinstance(kwargs, dict):
        as_list("x")
        as_list("y")
        data = pd.DataFrame(kwargs)
    else:
        data = kwargs

    mapping = aes()
    for ae in set(kwargs) & ALL_AESTHETICS:
        mapping[ae] = ae

    return geom_blank(data=data, mapping=mapping, inherit_aes=False)



from __future__ import annotations

import typing
from copy import deepcopy
from itertools import chain, repeat

import pandas as pd

from ..exceptions import PlotnineError
from ..layer import layer
from ..mapping.aes import is_valid_aesthetic, rename_aesthetics
from ..mapping.evaluation import evaluate
from ..positions.position import position
from ..stats.stat import stat
from ..utils import (
    Registry,
    copy_keys,
    data_mapping_as_kwargs,
    is_list_like,
    is_string,
    remove_missing,
)

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Aes,
        Axes,
        Coord,
        DataLike,
        DrawingArea,
        EvalEnvironment,
        Ggplot,
        Layer,
        Layout,
    )


class geom(metaclass=Registry):
    """Base class of all Geoms"""

    __base__ = True

    #: Default aesthetics for the geom
    DEFAULT_AES: dict[str, Any] = {}

    #: Required aesthetics for the geom
    REQUIRED_AES: set[str] = set()

    #: Required aesthetics for the geom
    NON_MISSING_AES: set[str] = set()

    #: Required parameters for the geom
    DEFAULT_PARAMS: dict[str, Any] = {}

    #: geom/layer specific dataframe
    data: DataLike

    #: mappings i.e. :py:`aes(x='col1', fill='col2')`
    mapping: Aes

    aes_params: dict[str, Any] = {}  # setting of aesthetic
    params: dict[str, Any]  # parameter settings

    # Plot namespace, it gets its value when the plot is being
    # built.
    environment: EvalEnvironment

    # The geom responsible for the legend if draw_legend is
    # not implemented
    legend_geom: str = "point"

    # Documentation for the aesthetics. It is added under the
    # documentation for mapping parameter. Use {aesthetics}
    # placeholder to insert a table for all the aesthetics and
    # their default values.
    _aesthetics_doc: str = "{aesthetics_table}"

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        kwargs = rename_aesthetics(kwargs)
        kwargs = data_mapping_as_kwargs((data, mapping), kwargs)
        self._kwargs = kwargs  # Will be used to create stat & layer

        # separate aesthetics and parameters
        self.aes_params = copy_keys(kwargs, {}, self.aesthetics())
        self.params = copy_keys(kwargs, deepcopy(self.DEFAULT_PARAMS))
        self.mapping = kwargs["mapping"]
        self.data = kwargs["data"]
        self._stat = stat.from_geom(self)
        self._position = position.from_geom(self)
        self._verify_arguments(kwargs)  # geom, stat, layer

    @staticmethod
    def from_stat(stat: stat) -> geom:
        """
        Return an instantiated geom object

        geoms should not override this method.

        Parameters
        ----------
        stat : stat
            `stat`

        Returns
        -------
        out : geom
            A geom object

        Raises
        ------
        PlotnineError
            If unable to create a `geom`.
        """
        name = stat.params["geom"]

        if isinstance(name, geom):
            return name

        if isinstance(name, type) and issubclass(name, geom):
            klass = name
        elif is_string(name):
            if not name.startswith("geom_"):
                name = f"geom_{name}"
            klass = Registry[name]
        else:
            raise PlotnineError(f"Unknown geom of type {type(name)}")

        return klass(stat=stat, **stat._kwargs)

    @classmethod
    def aesthetics(cls: type[geom]) -> set[str]:
        """
        Return all the aesthetics for this geom

        geoms should not override this method.
        """
        main = cls.DEFAULT_AES.keys() | cls.REQUIRED_AES
        other = {"group"}
        # Need to recognize both spellings
        if "color" in main:
            other.add("colour")
        if "outlier_color" in main:
            other.add("outlier_colour")
        return main | other

    def __deepcopy__(self, memo: dict[Any, Any]) -> geom:
        """
        Deep copy without copying the self.data dataframe

        geoms should not override this method.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a deepcopy of data, or environment
        shallow = {"data", "_kwargs", "environment"}
        for key, item in old.items():
            if key in shallow:
                new[key] = old[key]
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Modify the data before drawing takes place

        This function is called *before* position adjustments are done.
        It is used by geoms to create the final aesthetics used for
        drawing. The base class method does nothing, geoms can override
        this method for two reasons:

        1. The ``stat`` does not create all the aesthetics (usually
           position aesthetics) required for drawing the ``geom``,
           but those aesthetics can be computed from the available
           data. For example :class:`~plotnine.geoms.geom_boxplot`
           and :class:`~plotnine.geoms.geom_violin`.

        2. The ``geom`` inherits from another ``geom`` (superclass) which
           does the drawing and the superclass requires certain aesthetics
           to be present in the data. For example
           :class:`~plotnine.geoms.geom_tile` and
           :class:`~plotnine.geoms.geom_area`.

        Parameters
        ----------
        data : dataframe
            Data used for drawing the geom.

        Returns
        -------
        out : dataframe
            Data used for drawing the geom.
        """
        return data

    def use_defaults(
        self, data: pd.DataFrame, aes_modifiers: dict[str, Any]
    ) -> pd.DataFrame:
        """
        Combine data with defaults and set aesthetics from parameters

        geoms should not override this method.

        Parameters
        ----------
        data : dataframe
            Data used for drawing the geom.
        aes_modifiers : dict
            Aesthetics

        Returns
        -------
        out : dataframe
            Data used for drawing the geom.
        """
        missing_aes = (
            self.DEFAULT_AES.keys()
            - self.aes_params.keys()
            - set(data.columns.to_list())
        )

        # Not in data and not set, use default
        for ae in missing_aes:
            data[ae] = self.DEFAULT_AES[ae]

        # Evaluate/Modify the mapped aesthetics
        evaled = evaluate(aes_modifiers, data, self.environment)
        for ae in evaled.columns.intersection(data.columns):
            data[ae] = evaled[ae]

        if "PANEL" in data:
            num_panels = len(data["PANEL"].unique())
        else:
            num_panels = 1

        # Aesthetics set as parameters to the geom/stat
        for ae, value in self.aes_params.items():
            try:
                data[ae] = value
            except ValueError:
                # sniff out the special cases, like custom
                # tupled linetypes, shapes and colors
                if is_valid_aesthetic(value, ae):
                    data[ae] = [value] * len(data)
                elif num_panels > 1 and is_list_like(value):
                    data[ae] = list(chain(*repeat(value, num_panels)))
                else:
                    msg = "'{}' does not look like a " "valid value for `{}`"
                    raise PlotnineError(msg.format(value, ae))

        return data

    def draw_layer(
        self, data: pd.DataFrame, layout: Layout, coord: Coord, **params: Any
    ):
        """
        Draw layer across all panels

        geoms should not override this method.

        Parameters
        ----------
        data : DataFrame
            DataFrame specific for this layer
        layout : Layout
            Layout object created when the plot is getting
            built
        coord : coord
            Type of coordinate axes
        params : dict
            Combined *geom* and *stat* parameters. Also
            includes the stacking order of the layer in
            the plot (*zorder*)
        """
        for pid, pdata in data.groupby("PANEL"):
            if len(pdata) == 0:
                continue
            ploc = pdata["PANEL"].iat[0] - 1
            panel_params = layout.panel_params[ploc]
            ax = layout.axs[ploc]
            self.draw_panel(pdata, panel_params, coord, ax, **params)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups

        For efficiency, geoms that do not need to partition
        different groups before plotting should override this
        method and avoid the groupby.

        Parameters
        ----------
        data : dataframe
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params : panel_view
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Attributes are of interest
            to the geom are::

                'panel_params.x.range'  # tuple
                'panel_params.y.range'  # tuple

        coord : coord
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax : axes
            Axes on which to plot.
        params : dict
            Combined parameters for the geom and stat. Also
            includes the 'zorder'.
        """
        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True, drop=True)
            self.draw_group(gdata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot data belonging to a group.

        Parameters
        ----------
        data : dataframe
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params : panel_view
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Keys of interest to
            the geom are::

                'x_range'  # tuple
                'y_range'  # tuple

        coord : coord
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax : axes
            Axes on which to plot.
        params : dict
            Combined parameters for the geom and stat. Also
            includes the 'zorder'.
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)

    @staticmethod
    def draw_unit(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot data belonging to a unit.

        A matplotlib plot function may require that an aethestic
        have a single unique value. e.g. linestyle='dashed' and
        not linestyle=['dashed', 'dotted', ...].
        A single call to such a function can only plot lines with
        the same linestyle. However, if the plot we want has more
        than one line with different linestyles, we need to group
        the lines with the same linestyle and plot them as one
        unit. In this case, draw_group calls this function to do
        the plotting. For an example see
        :class:`~plotnine.geoms.geom_point`.

        Parameters
        ----------
        data : dataframe
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params : panel_view
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Keys of interest to
            the geom are::

                'x_range'  # tuple
                'y_range'  # tuple

            In rare cases a geom may need access to the x or y scales.
            Those are available at::

                'scales'   # SimpleNamespace

        coord : coord
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax : axes
            Axes on which to plot.
        params : dict
            Combined parameters for the geom and stat. Also
            includes the 'zorder'.
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)

    def __radd__(self, gg: Ggplot) -> Ggplot:
        """
        Add layer representing geom object on the right

        Parameters
        ----------
        gg : ggplot
            ggplot object

        Returns
        -------
        out : ggplot
            ggplot object with added layer.
        """
        gg += self.to_layer()  # Add layer
        return gg

    def to_layer(self) -> Layer:
        """
        Make a layer that represents this geom

        Returns
        -------
        out : layer
            Layer
        """
        return layer.from_geom(self)

    def _verify_arguments(self, kwargs: dict[str, Any]):
        """
        Verify arguments passed to the geom
        """
        geom_stat_args = kwargs.keys() | self._stat._kwargs.keys()
        unknown = (
            geom_stat_args
            - self.aesthetics()
            - self.DEFAULT_PARAMS.keys()  # geom aesthetics
            - self._stat.aesthetics()  # geom parameters
            - self._stat.DEFAULT_PARAMS.keys()  # stat aesthetics
            - {  # stat parameters
                "data",
                "mapping",
                "show_legend",  # layer parameters
                "inherit_aes",
                "raster",
            }
        )  # layer parameters
        if unknown:
            msg = (
                "Parameters {}, are not understood by "
                "either the geom, stat or layer."
            )
            raise PlotnineError(msg.format(unknown))

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Remove rows with NaN values

        geoms that infer extra information from missing values
        should override this method. For example
        :class:`~plotnine.geoms.geom_path`.

        Parameters
        ----------
        data : dataframe
            Data

        Returns
        -------
        out : dataframe
            Data without the NaNs.

        Notes
        -----
        Shows a warning if the any rows are removed and the
        `na_rm` parameter is False. It only takes into account
        the columns of the required aesthetics.
        """
        return remove_missing(
            data,
            self.params["na_rm"],
            list(self.REQUIRED_AES | self.NON_MISSING_AES),
            self.__class__.__name__,
        )

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        msg = "The geom should implement this method."
        raise NotImplementedError(msg)



from ..doctools import document
from .geom_path import geom_path

_params = geom_path.DEFAULT_PARAMS.copy()
_params["stat"] = "bin"


@document
class geom_freqpoly(geom_path):
    """
    Frequency polygon

    {usage}

    See :class:`~plotnine.geoms.geom_path` for documentation
    of the parameters.
    """

    DEFAULT_PARAMS = _params



from __future__ import annotations

import typing

from ..doctools import document
from .geom_ribbon import geom_ribbon

if typing.TYPE_CHECKING:
    import pandas as pd


@document
class geom_area(geom_ribbon):
    """
    Area plot

    An area plot is a special case of geom_ribbon,
    where the minimum of the range is fixed to 0,
    and the position adjustment defaults to 'stack'.

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_ribbon
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = dict(
        geom_ribbon.DEFAULT_PARAMS,
        **{"position": "stack", "outline_type": "upper"},
    )

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        data["ymin"] = 0
        data["ymax"] = data["y"]
        return data



from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_datetime, scale_discrete


@document
class scale_alpha(scale_continuous):
    """
    Continuous Alpha Scale

    Parameters
    ----------
    range : array_like
        Range ([Minimum, Maximum]) of output alpha values.
        Should be between 0 and 1. Default is ``(0.1, 1)``
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]

    def __init__(self, range=(0.1, 1), **kwargs):
        from mizani.palettes import rescale_pal

        # TODO: fix types in mizani
        self.palette = rescale_pal(range)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


alias("scale_alpha_continuous", scale_alpha)


@document
class scale_alpha_ordinal(scale_discrete):
    """
    Ordinal Alpha Scale

    Parameters
    ----------
    range : array_like
        Range ([Minimum, Maximum]) of output alpha values.
        Should be between 0 and 1. Default is ``(0.1, 1)``
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]

    def __init__(self, range=(0.1, 1), **kwargs):
        def palette(n):
            return np.linspace(range[0], range[1], n)

        self.palette = palette
        scale_discrete.__init__(self, **kwargs)


@document
class scale_alpha_discrete(scale_alpha_ordinal):
    """
    Discrete Alpha Scale

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]

    def __init__(self, **kwargs):
        warn(
            "Using alpha for a discrete variable is not advised.",
            PlotnineWarning,
        )
        super().__init__(**kwargs)


@document
class scale_alpha_datetime(scale_datetime):
    """
    Datetime Alpha Scale

    Parameters
    ----------
    range : array_like
        Range ([Minimum, Maximum]) of output alpha values.
        Should be between 0 and 1. Default is ``(0.1, 1)``
    {superclass_parameters}
    """

    _aesthetics = ["alpha"]

    def __init__(self, range=(0.1, 1), **kwargs):
        from mizani.palettes import rescale_pal

        # TODO: fix types in mizani
        self.palette = rescale_pal(range)  # pyright: ignore
        scale_datetime.__init__(self, **kwargs)



from ..options import get_option
from ..utils import alias
from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme


class theme_gray(theme):
    """
    A gray background with white gridlines.

    This is the default theme

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        base_family = base_family or get_option("base_family")
        quarter_line = base_size / 4
        fifth_line = base_size / 5
        eigth_line = base_size / 8
        m = get_option("base_margin")

        super().__init__(
            line=element_line(
                color="black", size=1, linetype="solid", lineend="butt"
            ),
            rect=element_rect(
                fill="white", color="black", size=1, linetype="solid"
            ),
            text=element_text(
                family=base_family,
                style="normal",
                color="black",
                size=base_size,
                linespacing=0.9,
                ha="center",
                va="center",
                rotation=0,
                margin={},
            ),
            aspect_ratio=get_option("aspect_ratio"),
            axis_line=element_line(),
            axis_line_x=element_blank(),
            axis_line_y=element_blank(),
            axis_text=element_text(size=base_size * 0.8, color="#4D4D4D"),
            axis_text_x=element_text(va="top", margin={"t": fifth_line}),
            axis_text_y=element_text(ha="right", margin={"r": fifth_line}),
            axis_ticks=element_line(color="#333333"),
            axis_ticks_direction="out",
            axis_ticks_length=0,
            axis_ticks_length_major=quarter_line,
            axis_ticks_length_minor=eigth_line,
            axis_ticks_minor=element_blank(),
            axis_ticks_pad=2,
            axis_title_x=element_text(
                va="bottom", ha="center", margin={"t": m, "units": "fig"}
            ),
            axis_title_y=element_text(
                angle=90,
                va="center",
                ha="left",
                margin={"r": m, "units": "fig"},
            ),
            dpi=get_option("dpi"),
            figure_size=get_option("figure_size"),
            # legend, None values are for parameters where the
            # drawing routines can make better decisions than
            # can be pre-determined in the theme.
            legend_background=element_rect(color="None"),
            legend_box="auto",
            legend_box_just="auto",
            legend_box_margin=0,  # points
            legend_box_spacing=m * 3,  # figure units
            legend_direction="auto",
            legend_entry_spacing_x=5,
            legend_entry_spacing_y=2,
            legend_key=element_rect(fill="#F2F2F2", colour="None"),
            legend_key_height=None,
            legend_key_size=base_size * 0.8 * 1.8,
            legend_key_width=None,
            legend_margin=0,  # points
            legend_position="right",
            legend_spacing=10,  # points
            legend_text=element_text(
                size=base_size * 0.8,
                ha="left",
                margin={"t": 3, "b": 3, "l": 3, "r": 3, "units": "pt"},
            ),
            legend_text_colorbar=element_text(va="center"),
            legend_text_legend=element_text(va="baseline"),
            legend_title=element_text(
                ha="left",
                margin={
                    "t": m,
                    "b": m,
                    "l": m,
                    "r": m,
                    "units": "fig",
                },
            ),
            legend_title_align="auto",
            panel_background=element_rect(fill="#EBEBEB"),
            panel_border=element_blank(),
            panel_grid_major=element_line(color="white", size=1),
            panel_grid_minor=element_line(color="white", size=0.5),
            panel_spacing=m,
            panel_spacing_x=None,
            panel_spacing_y=None,
            plot_background=element_rect(color="white"),
            plot_caption=element_text(
                size=base_size * 0.8,
                ha="right",
                va="bottom",
                ma="left",
                margin={"t": m, "units": "fig"},
            ),
            plot_margin=m,
            plot_subtitle=element_text(
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            plot_title=element_text(
                size=base_size * 1.2,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            strip_align=0,
            strip_align_x=None,
            strip_align_y=None,
            strip_background=element_rect(color="None", fill="#D9D9D9"),
            strip_background_x=element_rect(width=1),
            strip_background_y=element_rect(height=1),
            strip_text=element_text(
                color="#1A1A1A",
                size=base_size * 0.8,
                linespacing=1.0,
                margin={
                    "t": 1 / 3,
                    "b": 1 / 3,
                    "l": 1 / 3,
                    "r": 1 / 3,
                    "units": "lines",
                },
            ),
            strip_text_x=None,
            strip_text_y=element_text(rotation=-90),
            complete=True,
        )


alias("theme_grey", theme_gray)



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, geom_raster, ggplot
from plotnine.exceptions import PlotnineWarning


def _random_grid(n, m=None, seed=123):
    if m is None:
        m = n
    prg = np.random.RandomState(seed)
    g = prg.uniform(size=n * m)
    x, y = np.meshgrid(range(n), range(m))
    df = pd.DataFrame({"x": x.ravel(), "y": y.ravel(), "g": g})
    return df


def test_square():
    df = _random_grid(5)
    p = ggplot(df, aes("x", "y", fill="g")) + geom_raster(
        interpolation="bilinear"
    )
    assert p == "square"


def test_rectangle():
    df = _random_grid(3, 5)
    p = ggplot(df, aes("x", "y", fill="g")) + geom_raster(
        interpolation="bilinear"
    )
    assert p == "rectangle"


def test_gap_no_interpolation():
    df = _random_grid(3, 2)
    df.at[4, "y"] = 3
    p = ggplot(df, aes("x", "y", fill="g")) + geom_raster()
    # Warns about uneven vertical intervals
    with pytest.warns(PlotnineWarning):
        assert p == "gap_no_interpolation"


def test_gap_with_interpolation():
    df = _random_grid(3, 2)
    df.at[4, "y"] = 3
    p = ggplot(df, aes("x", "y", fill="g")) + geom_raster(
        interpolation="bilinear"
    )
    # Warns about uneven vertical intervals
    with pytest.warns(PlotnineWarning):
        assert p == "gap_with_interpolation"



from ..doctools import document
from .geom_path import geom_path


@document
class geom_quantile(geom_path):
    """
    Quantile lines from a quantile regression

    {usage}

    Parameters
    ----------
    {common_parameters}
    lineend : str (default: butt)
        Line end style, of of *butt*, *round* or *projecting.*
        This option is applied for solid linetypes.
    linejoin : str (default: round)
        Line join style, one of *round*, *miter* or *bevel*.
        This option is applied for solid linetypes.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "#3366FF",
        "linetype": "solid",
        "size": 0.5,
    }
    DEFAULT_PARAMS = {
        "stat": "quantile",
        "position": "identity",
        "na_rm": False,
        "lineend": "butt",
        "linejoin": "round",
    }



from .position import position


class position_identity(position):
    """
    Do not adjust the position
    """

    @classmethod
    def compute_layer(cls, data, params, layout):
        return data



import numpy as np
import pandas as pd

from plotnine import aes, geom_spoke, ggplot

n = 4
df = pd.DataFrame(
    {
        "x": [1] * n,
        "y": range(n),
        "angle": np.linspace(0, np.pi / 2, n),
        "radius": range(1, n + 1),
        "z": range(n),
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes(y="y", angle="angle", radius="radius"))
        + geom_spoke(aes("x"), size=2)
        + geom_spoke(aes("x+2", alpha="z"), size=2)
        + geom_spoke(aes("x+4", linetype="factor(z)"), size=2)
        + geom_spoke(aes("x+6", color="factor(z)"), size=2)
        + geom_spoke(aes("x+8", size="z"))
    )

    assert p == "aesthetics"


def test_unmapped_angle():
    p = ggplot(df, aes(y="y", angle="angle", radius="radius")) + geom_spoke(
        aes("x", "y"), angle=0, radius=1
    )
    assert p == "test_unmapped_angle"



import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import pytest

from plotnine import (
    aes,
    facet_wrap,
    geom_point,
    geom_text,
    ggplot,
    ggsave,
    theme_xkcd,
)
from plotnine.data import mtcars
from plotnine.exceptions import PlotnineError, PlotnineWarning

p = ggplot(mtcars, aes(x="wt", y="mpg", label="name")) + geom_text()


def sequential_filenames():
    """
    Generate filenames for the tests
    """
    for i in range(100):
        yield Path(f"filename-{i}.png")


filename_gen = sequential_filenames()


def assert_exist_and_clean(filename, msg=None):
    if not msg:
        msg = f"File {filename} does not exist"
    assert filename.exists(), msg
    filename.unlink()


class TestArguments:
    def test_default_filename(self):
        p.save(verbose=False)
        fn = p._save_filename("pdf")
        assert_exist_and_clean(fn, "default filename")

    def test_save_method(self):
        fn = next(filename_gen)
        with pytest.warns(PlotnineWarning) as record:
            p.save(fn)

        assert_exist_and_clean(fn, "save method")

        res = ("saving" in str(item.message).lower() for item in record)
        assert any(res)

        res = ("filename" in str(item.message).lower() for item in record)
        assert any(res)

        # verbose
        fn = next(filename_gen)
        with warnings.catch_warnings(record=True) as record:
            p.save(fn, verbose=False)
            assert_exist_and_clean(fn, "save method")
            assert not record, "Issued an unexpected warning"

    def test_filename_plot_path(self):
        fn = next(filename_gen)
        p.save(fn, path=".", verbose=False)
        assert_exist_and_clean(fn, "fn, plot and path")

    def test_format_png(self):
        p.save(format="png", verbose=False)
        fn = p._save_filename("png")
        assert_exist_and_clean(fn, "format png")

    def test_dpi(self):
        fn = next(filename_gen)
        p.save(fn, dpi=100, verbose=False)
        assert_exist_and_clean(fn, "dpi = 100")

    def test_ggsave(self):
        ggsave(p, verbose=False)
        fn = p._save_filename("pdf")
        assert_exist_and_clean(fn, "default filename")

    def test_save_big(self):
        fn = next(filename_gen)
        # supplying the ggplot object will work without
        # printing it first! 26 is the current limit, just go
        # over it to not use too much memory
        p.save(fn, width=26, height=26, limitsize=False, verbose=False)
        assert_exist_and_clean(fn, "big height and width")

    def test_dpi_theme_xkcd(self):
        fn1 = next(filename_gen)
        fn2 = next(filename_gen)

        df = pd.DataFrame({"x": range(4), "y": range(4), "b": list("aabb")})

        p = (
            ggplot(df)
            + geom_point(aes("x", "y"))
            + facet_wrap("b")
            + theme_xkcd()
        )
        p.save(fn1, verbose=False)
        assert_exist_and_clean(fn1, "Saving with theme_xkcd and dpi (1)")

        p.save(fn2, dpi=72, verbose=False)
        assert_exist_and_clean(fn2, "Saving with theme_xkcd and dpi (2)")


class TestExceptions:
    def test_unknown_format(self):
        with pytest.raises(Exception):
            p.save(format="unknown", verbose=False)

    def test_width_only(self):
        with pytest.raises(PlotnineError):
            p.save(width=11)

    def test_height_only(self):
        with pytest.raises(PlotnineError):
            p.save(height=8)

    def test_large_width(self):
        with pytest.raises(PlotnineError):
            p.save(width=300, height=8)

    def test_large_height(self):
        with pytest.raises(PlotnineError):
            p.save(widhth=11, height=300)

    def test_bad_units(self):
        with pytest.raises(Exception):
            p.save(width=1, heigth=1, units="xxx")


# This should be the last function in the file since it can catch
# "leakages" due to the tests in this test module.
def test_ggsave_closes_plot():
    assert plt.get_fignums() == [], "There are unsaved test plots"
    fn = next(filename_gen)
    p.save(fn, verbose=False)
    assert_exist_and_clean(fn, "exist")
    assert plt.get_fignums() == [], "ggplot.save did not close the plot"



"""
Kernel Density Functions

These functions make it easy to integrate stats that compute
kernel densities with the wider scientific python ecosystem.

Credit: Jake VanderPlas for the original kde_* functions
https://jakevdp.github.io/blog/2013/12/01/kernel-density-estimation/
"""

import numpy as np
import pandas.api.types as pdtypes


def kde_scipy(data, grid, **kwargs):
    """
    Kernel Density Estimation with Scipy

    Parameters
    ----------
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x p` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x p` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    from scipy.stats import gaussian_kde

    kde = gaussian_kde(data.T, **kwargs)
    return kde.evaluate(grid.T)


def kde_statsmodels_u(data, grid, **kwargs):
    """
    Univariate Kernel Density Estimation with Statsmodels

    Parameters
    ----------
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x 1` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x 1` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    from statsmodels.nonparametric.kde import KDEUnivariate

    kde = KDEUnivariate(data)
    kde.fit(**kwargs)
    return kde.evaluate(grid)


def kde_statsmodels_m(data, grid, **kwargs):
    """
    Multivariate Kernel Density Estimation with Statsmodels

    Parameters
    ----------
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x p` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x p` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    from statsmodels.nonparametric.kernel_density import KDEMultivariate

    kde = KDEMultivariate(data, **kwargs)
    return kde.pdf(grid)


def kde_sklearn(data, grid, **kwargs):
    """
    Kernel Density Estimation with Scikit-learn

    Parameters
    ----------
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x p` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x p` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    # Not core dependency
    try:
        from sklearn.neighbors import KernelDensity
    except ImportError as err:
        raise ImportError("scikit-learn is not installed") from err
    kde_skl = KernelDensity(**kwargs)
    kde_skl.fit(data)
    # score_samples() returns the log-likelihood of the samples
    log_pdf = kde_skl.score_samples(grid)
    return np.exp(log_pdf)


def kde_count(data, grid, **kwargs):
    """
    Kernel Density Estimation via count within radius

    Parameters
    ----------
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x p` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x p` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    r = kwargs.get("radius", np.ptp(data) / 10)

    # Get the number of data points within the radius r of each grid point
    iter = (np.sum(np.linalg.norm(data - g, axis=1) < r) for g in grid)
    count = np.fromiter(iter, float, count=data.shape[0])

    # Get fraction of data within radius
    density = count / data.shape[0]

    return density


KDE_FUNCS = {
    "statsmodels-u": kde_statsmodels_u,
    "statsmodels-m": kde_statsmodels_m,
    "scipy": kde_scipy,
    "scikit-learn": kde_sklearn,
    "sklearn": kde_sklearn,
    "count": kde_count,
}


def kde(data, grid, package, **kwargs):
    """
    Kernel Density Estimation

    Parameters
    ----------
    package : str
        Package whose kernel density estimation to use.
        Should be one of
        `['statsmodels-u', 'statsmodels-m', 'scipy', 'sklearn']`.
    data : numpy.array
        Data points used to compute a density estimator. It
        has `n x p` dimensions, representing n points and p
        variables.
    grid : numpy.array
        Data points at which the desity will be estimated. It
        has `m x p` dimensions, representing m points and p
        variables.

    Returns
    -------
    out : numpy.array
        Density estimate. Has `m x 1` dimensions
    """
    if package == "statsmodels":
        package = "statsmodels-m"
    func = KDE_FUNCS[package]
    return func(data, grid, **kwargs)


def get_var_type(col):
    """
    Return var_type (for KDEMultivariate) of the column

    Parameters
    ----------
    col : pandas.Series
        A dataframe column.

    Returns
    -------
    out : str
        One of ['c', 'o', 'u'].

    See Also
    --------
    The origin of the character codes is
    :class:`statsmodels.nonparametric.kernel_density.KDEMultivariate`.
    """
    if pdtypes.is_numeric_dtype(col):
        # continuous
        return "c"
    elif pdtypes.is_categorical_dtype(col):
        # ordered or unordered
        return "o" if col.cat.ordered else "u"
    else:
        # unordered if unsure, e.g string columns that
        # are not categorical
        return "u"
